# This file is automatically generated by botocraft.  Do not edit directly.
# mypy: disable-error-code="index, override, assignment, union-attr, misc"
from botocraft.mixins.tags import TagsDictMixin
from botocraft.mixins.application_autoscaling import ScalableTargetModelMixin
from .abstract import (
    Boto3Model,
    ReadonlyBoto3Model,
    PrimaryBoto3Model,
    ReadonlyPrimaryBoto3Model,
    Boto3ModelManager,
    ReadonlyBoto3ModelManager,
)
from typing import ClassVar, Type, Optional, Literal, Dict, List, Union, Any, cast
from functools import cached_property
from .abstract import PrimaryBoto3ModelQuerySet
from botocraft.mixins.application_autoscaling import scaling_policy_only
from datetime import datetime
from collections import OrderedDict
from pydantic import Field
from botocraft.mixins.application_autoscaling import scalable_target_only

# ===============
# Managers
# ===============


class ScalingPolicyManager(Boto3ModelManager):
    service_name: str = "application-autoscaling"

    @scaling_policy_only
    def create(self, model: "ScalingPolicy") -> "PutScalingPolicyResponse":
        """
        Creates or updates a scaling policy for an Application Auto Scaling scalable
        target.

        Args:
            model: The :py:class:`ScalingPolicy` to create.
        """
        data = model.model_dump(exclude_none=True, by_alias=True)
        args = dict(
            PolicyName=data.get("PolicyName"),
            ServiceNamespace=data.get("ServiceNamespace"),
            ResourceId=data.get("ResourceId"),
            ScalableDimension=data.get("ScalableDimension"),
            PolicyType=data.get("PolicyType"),
            StepScalingPolicyConfiguration=data.get("StepScalingPolicyConfiguration"),
            TargetTrackingScalingPolicyConfiguration=data.get(
                "TargetTrackingScalingPolicyConfiguration"
            ),
            PredictiveScalingPolicyConfiguration=data.get(
                "PredictiveScalingPolicyConfiguration"
            ),
        )
        _response = self.client.put_scaling_policy(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = PutScalingPolicyResponse(**_response)

        self.sessionize(response)
        return cast("PutScalingPolicyResponse", response)

    def delete(
        self,
        PolicyName: str,
        ServiceNamespace: Literal[
            "ecs",
            "elasticmapreduce",
            "ec2",
            "appstream",
            "dynamodb",
            "rds",
            "sagemaker",
            "custom-resource",
            "comprehend",
            "lambda",
            "cassandra",
            "kafka",
            "elasticache",
            "neptune",
            "workspaces",
        ],
        ResourceId: str,
        ScalableDimension: Literal[
            "ecs:service:DesiredCount",
            "ec2:spot-fleet-request:TargetCapacity",
            "elasticmapreduce:instancegroup:InstanceCount",
            "appstream:fleet:DesiredCapacity",
            "dynamodb:table:ReadCapacityUnits",
            "dynamodb:table:WriteCapacityUnits",
            "dynamodb:index:ReadCapacityUnits",
            "dynamodb:index:WriteCapacityUnits",
            "rds:cluster:ReadReplicaCount",
            "sagemaker:variant:DesiredInstanceCount",
            "custom-resource:ResourceType:Property",
            "comprehend:document-classifier-endpoint:DesiredInferenceUnits",
            "comprehend:entity-recognizer-endpoint:DesiredInferenceUnits",
            "lambda:function:ProvisionedConcurrency",
            "cassandra:table:ReadCapacityUnits",
            "cassandra:table:WriteCapacityUnits",
            "kafka:broker-storage:VolumeSize",
            "elasticache:cache-cluster:Nodes",
            "elasticache:replication-group:NodeGroups",
            "elasticache:replication-group:Replicas",
            "neptune:cluster:ReadReplicaCount",
            "sagemaker:variant:DesiredProvisionedConcurrency",
            "sagemaker:inference-component:DesiredCopyCount",
            "workspaces:workspacespool:DesiredUserSessions",
        ],
    ) -> None:
        """
        Deletes the specified scaling policy for an Application Auto Scaling scalable
        target.

        Args:
            PolicyName: The name of the scaling policy.
            ServiceNamespace: The namespace of the Amazon Web Services service that provides the resource. For a resource
                provided by your own application or service, use ``custom-resource`` instead.
            ResourceId: The identifier of the resource associated with the scalable target. This string consists of the resource
                type and unique identifier.
            ScalableDimension: The scalable dimension. This string consists of the service namespace, resource type, and scaling
                property.
        """
        args: Dict[str, Any] = dict(
            PolicyName=self.serialize(PolicyName),
            ServiceNamespace=self.serialize(ServiceNamespace),
            ResourceId=self.serialize(ResourceId),
            ScalableDimension=self.serialize(ScalableDimension),
        )
        self.client.delete_scaling_policy(
            **{k: v for k, v in args.items() if v is not None}
        )

    def list(
        self,
        ServiceNamespace: Literal[
            "ecs",
            "elasticmapreduce",
            "ec2",
            "appstream",
            "dynamodb",
            "rds",
            "sagemaker",
            "custom-resource",
            "comprehend",
            "lambda",
            "cassandra",
            "kafka",
            "elasticache",
            "neptune",
            "workspaces",
        ],
        *,
        PolicyNames: List[str] | None = None,
        ResourceId: str | None = None,
        ScalableDimension: Literal["ecs:service:DesiredCount", "ec2:spot-fleet-request:TargetCapacity", "elasticmapreduce:instancegroup:InstanceCount", "appstream:fleet:DesiredCapacity", "dynamodb:table:ReadCapacityUnits", "dynamodb:table:WriteCapacityUnits", "dynamodb:index:ReadCapacityUnits", "dynamodb:index:WriteCapacityUnits", "rds:cluster:ReadReplicaCount", "sagemaker:variant:DesiredInstanceCount", "custom-resource:ResourceType:Property", "comprehend:document-classifier-endpoint:DesiredInferenceUnits", "comprehend:entity-recognizer-endpoint:DesiredInferenceUnits", "lambda:function:ProvisionedConcurrency", "cassandra:table:ReadCapacityUnits", "cassandra:table:WriteCapacityUnits", "kafka:broker-storage:VolumeSize", "elasticache:cache-cluster:Nodes", "elasticache:replication-group:NodeGroups", "elasticache:replication-group:Replicas", "neptune:cluster:ReadReplicaCount", "sagemaker:variant:DesiredProvisionedConcurrency", "sagemaker:inference-component:DesiredCopyCount", "workspaces:workspacespool:DesiredUserSessions"] | None = None,
    ) -> PrimaryBoto3ModelQuerySet:
        """
        Describes the Application Auto Scaling scaling policies for the specified
        service namespace.

        Args:
            ServiceNamespace: The namespace of the Amazon Web Services service that provides the resource. For a resource
                provided by your own application or service, use ``custom-resource`` instead.

        Keyword Args:
            PolicyNames: The names of the scaling policies to describe.
            ResourceId: The identifier of the resource associated with the scaling policy. This string consists of the resource
                type and unique identifier.
            ScalableDimension: The scalable dimension. This string consists of the service namespace, resource type, and scaling
                property. If you specify a scalable dimension, you must also specify a resource ID.
        """
        paginator = self.client.get_paginator("describe_scaling_policies")
        args: Dict[str, Any] = dict(
            ServiceNamespace=self.serialize(ServiceNamespace),
            PolicyNames=self.serialize(PolicyNames),
            ResourceId=self.serialize(ResourceId),
            ScalableDimension=self.serialize(ScalableDimension),
        )
        response_iterator = paginator.paginate(
            **{k: v for k, v in args.items() if v is not None}
        )
        results = []
        for _response in response_iterator:
            if list(_response.keys()) == ["ResponseMetadata"]:
                break
            if "ResponseMetadata" in _response:
                del _response["ResponseMetadata"]
            response = DescribeScalingPoliciesResponse(**_response)
            if response.ScalingPolicies:
                results.extend(response.ScalingPolicies)
            else:
                if getattr(response, "NextToken", None):
                    continue
                break
        self.sessionize(results)
        if results and isinstance(results[0], Boto3Model):
            return PrimaryBoto3ModelQuerySet(results)
        return results

    @scaling_policy_only
    def update(self, model: "ScalingPolicy") -> "PutScalingPolicyResponse":
        """
        Creates or updates a scaling policy for an Application Auto Scaling scalable
        target.

        Args:
            model: The :py:class:`ScalingPolicy` to update.
        """
        data = model.model_dump(exclude_none=True, by_alias=True)
        args = dict(
            PolicyName=data.get("PolicyName"),
            ServiceNamespace=data.get("ServiceNamespace"),
            ResourceId=data.get("ResourceId"),
            ScalableDimension=data.get("ScalableDimension"),
            PolicyType=data.get("PolicyType"),
            StepScalingPolicyConfiguration=data.get("StepScalingPolicyConfiguration"),
            TargetTrackingScalingPolicyConfiguration=data.get(
                "TargetTrackingScalingPolicyConfiguration"
            ),
            PredictiveScalingPolicyConfiguration=data.get(
                "PredictiveScalingPolicyConfiguration"
            ),
        )
        _response = self.client.put_scaling_policy(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = PutScalingPolicyResponse(**_response)

        self.sessionize(response)
        return cast("PutScalingPolicyResponse", response)

    def get(
        self,
        ServiceNamespace: Literal[
            "ecs",
            "elasticmapreduce",
            "ec2",
            "appstream",
            "dynamodb",
            "rds",
            "sagemaker",
            "custom-resource",
            "comprehend",
            "lambda",
            "cassandra",
            "kafka",
            "elasticache",
            "neptune",
            "workspaces",
        ],
        *,
        PolicyNames: List[str] | None = None,
        ResourceId: str | None = None,
        ScalableDimension: Literal["ecs:service:DesiredCount", "ec2:spot-fleet-request:TargetCapacity", "elasticmapreduce:instancegroup:InstanceCount", "appstream:fleet:DesiredCapacity", "dynamodb:table:ReadCapacityUnits", "dynamodb:table:WriteCapacityUnits", "dynamodb:index:ReadCapacityUnits", "dynamodb:index:WriteCapacityUnits", "rds:cluster:ReadReplicaCount", "sagemaker:variant:DesiredInstanceCount", "custom-resource:ResourceType:Property", "comprehend:document-classifier-endpoint:DesiredInferenceUnits", "comprehend:entity-recognizer-endpoint:DesiredInferenceUnits", "lambda:function:ProvisionedConcurrency", "cassandra:table:ReadCapacityUnits", "cassandra:table:WriteCapacityUnits", "kafka:broker-storage:VolumeSize", "elasticache:cache-cluster:Nodes", "elasticache:replication-group:NodeGroups", "elasticache:replication-group:Replicas", "neptune:cluster:ReadReplicaCount", "sagemaker:variant:DesiredProvisionedConcurrency", "sagemaker:inference-component:DesiredCopyCount", "workspaces:workspacespool:DesiredUserSessions"] | None = None,
    ) -> Optional["ScalingPolicy"]:
        """
        Describes the Application Auto Scaling scaling policies for the specified
        service namespace.

        Args:
            ServiceNamespace: The namespace of the Amazon Web Services service that provides the resource. For a resource
                provided by your own application or service, use ``custom-resource`` instead.

        Keyword Args:
            PolicyNames: The names of the scaling policies to describe.
            ResourceId: The identifier of the resource associated with the scaling policy. This string consists of the resource
                type and unique identifier.
            ScalableDimension: The scalable dimension. This string consists of the service namespace, resource type, and scaling
                property. If you specify a scalable dimension, you must also specify a resource ID.
        """
        args: Dict[str, Any] = dict(
            ServiceNamespace=self.serialize(ServiceNamespace),
            PolicyNames=self.serialize(PolicyNames),
            ResourceId=self.serialize(ResourceId),
            ScalableDimension=self.serialize(ScalableDimension),
        )
        _response = self.client.describe_scaling_policies(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = DescribeScalingPoliciesResponse(**_response)

        if response and response.ScalingPolicies:
            self.sessionize(response.ScalingPolicies[0])
            return response.ScalingPolicies[0]
        return None


class ScalableTargetManager(Boto3ModelManager):
    service_name: str = "application-autoscaling"

    @scalable_target_only
    def create(
        self, model: "ScalableTarget", Tags: Dict[str, str] | None = None
    ) -> "RegisterScalableTargetResponse":
        """
        Registers or updates a scalable target, which is the resource that you want to
        scale.

        Args:
            model: The :py:class:`ScalableTarget` to create.

        Keyword Args:
            Tags: Assigns one or more tags to the scalable target. Use this parameter to tag the scalable target when it is
                created. To tag an existing scalable target, use the TagResource operation.
        """
        data = model.model_dump(exclude_none=True, by_alias=True)
        args = dict(
            ServiceNamespace=data.get("ServiceNamespace"),
            ResourceId=data.get("ResourceId"),
            ScalableDimension=data.get("ScalableDimension"),
            MinCapacity=data.get("MinCapacity"),
            MaxCapacity=data.get("MaxCapacity"),
            RoleARN=data.get("RoleARN"),
            SuspendedState=data.get("SuspendedState"),
            Tags=self.serialize(Tags),
        )
        _response = self.client.register_scalable_target(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = RegisterScalableTargetResponse(**_response)

        self.sessionize(response)
        return cast("RegisterScalableTargetResponse", response)

    def delete(
        self,
        ServiceNamespace: Literal[
            "ecs",
            "elasticmapreduce",
            "ec2",
            "appstream",
            "dynamodb",
            "rds",
            "sagemaker",
            "custom-resource",
            "comprehend",
            "lambda",
            "cassandra",
            "kafka",
            "elasticache",
            "neptune",
            "workspaces",
        ],
        ResourceId: str,
        ScalableDimension: Literal[
            "ecs:service:DesiredCount",
            "ec2:spot-fleet-request:TargetCapacity",
            "elasticmapreduce:instancegroup:InstanceCount",
            "appstream:fleet:DesiredCapacity",
            "dynamodb:table:ReadCapacityUnits",
            "dynamodb:table:WriteCapacityUnits",
            "dynamodb:index:ReadCapacityUnits",
            "dynamodb:index:WriteCapacityUnits",
            "rds:cluster:ReadReplicaCount",
            "sagemaker:variant:DesiredInstanceCount",
            "custom-resource:ResourceType:Property",
            "comprehend:document-classifier-endpoint:DesiredInferenceUnits",
            "comprehend:entity-recognizer-endpoint:DesiredInferenceUnits",
            "lambda:function:ProvisionedConcurrency",
            "cassandra:table:ReadCapacityUnits",
            "cassandra:table:WriteCapacityUnits",
            "kafka:broker-storage:VolumeSize",
            "elasticache:cache-cluster:Nodes",
            "elasticache:replication-group:NodeGroups",
            "elasticache:replication-group:Replicas",
            "neptune:cluster:ReadReplicaCount",
            "sagemaker:variant:DesiredProvisionedConcurrency",
            "sagemaker:inference-component:DesiredCopyCount",
            "workspaces:workspacespool:DesiredUserSessions",
        ],
    ) -> None:
        """
        Deregisters an Application Auto Scaling scalable target when you have finished using it. To see which resources have
        been registered, use `DescribeScalableTargets <https://docs.aws.amazon.com/autoscaling/application/APIReference/API_Desc
        ribeScalableTargets.html>`_.

        Args:
            ServiceNamespace: The namespace of the Amazon Web Services service that provides the resource. For a resource
                provided by your own application or service, use ``custom-resource`` instead.
            ResourceId: The identifier of the resource associated with the scalable target. This string consists of the resource
                type and unique identifier.
            ScalableDimension: The scalable dimension associated with the scalable target. This string consists of the service
                namespace, resource type, and scaling property.

        """
        args: Dict[str, Any] = dict(
            ServiceNamespace=self.serialize(ServiceNamespace),
            ResourceId=self.serialize(ResourceId),
            ScalableDimension=self.serialize(ScalableDimension),
        )
        self.client.deregister_scalable_target(
            **{k: v for k, v in args.items() if v is not None}
        )

    def list(
        self,
        ServiceNamespace: Literal[
            "ecs",
            "elasticmapreduce",
            "ec2",
            "appstream",
            "dynamodb",
            "rds",
            "sagemaker",
            "custom-resource",
            "comprehend",
            "lambda",
            "cassandra",
            "kafka",
            "elasticache",
            "neptune",
            "workspaces",
        ],
        *,
        ResourceIds: List[str] | None = None,
        ScalableDimension: Literal["ecs:service:DesiredCount", "ec2:spot-fleet-request:TargetCapacity", "elasticmapreduce:instancegroup:InstanceCount", "appstream:fleet:DesiredCapacity", "dynamodb:table:ReadCapacityUnits", "dynamodb:table:WriteCapacityUnits", "dynamodb:index:ReadCapacityUnits", "dynamodb:index:WriteCapacityUnits", "rds:cluster:ReadReplicaCount", "sagemaker:variant:DesiredInstanceCount", "custom-resource:ResourceType:Property", "comprehend:document-classifier-endpoint:DesiredInferenceUnits", "comprehend:entity-recognizer-endpoint:DesiredInferenceUnits", "lambda:function:ProvisionedConcurrency", "cassandra:table:ReadCapacityUnits", "cassandra:table:WriteCapacityUnits", "kafka:broker-storage:VolumeSize", "elasticache:cache-cluster:Nodes", "elasticache:replication-group:NodeGroups", "elasticache:replication-group:Replicas", "neptune:cluster:ReadReplicaCount", "sagemaker:variant:DesiredProvisionedConcurrency", "sagemaker:inference-component:DesiredCopyCount", "workspaces:workspacespool:DesiredUserSessions"] | None = None,
    ) -> PrimaryBoto3ModelQuerySet:
        """
        Gets information about the scalable targets in the specified namespace.

        Args:
            ServiceNamespace: The namespace of the Amazon Web Services service that provides the resource. For a resource
                provided by your own application or service, use ``custom-resource`` instead.

        Keyword Args:
            ResourceIds: The identifier of the resource associated with the scalable target. This string consists of the
                resource type and unique identifier.
            ScalableDimension: The scalable dimension associated with the scalable target. This string consists of the service
                namespace, resource type, and scaling property. If you specify a scalable dimension, you must also specify a
                resource ID.
        """
        paginator = self.client.get_paginator("describe_scalable_targets")
        args: Dict[str, Any] = dict(
            ServiceNamespace=self.serialize(ServiceNamespace),
            ResourceIds=self.serialize(ResourceIds),
            ScalableDimension=self.serialize(ScalableDimension),
        )
        response_iterator = paginator.paginate(
            **{k: v for k, v in args.items() if v is not None}
        )
        results = []
        for _response in response_iterator:
            if list(_response.keys()) == ["ResponseMetadata"]:
                break
            if "ResponseMetadata" in _response:
                del _response["ResponseMetadata"]
            response = DescribeScalableTargetsResponse(**_response)
            if response.ScalableTargets:
                results.extend(response.ScalableTargets)
            else:
                if getattr(response, "NextToken", None):
                    continue
                break
        self.sessionize(results)
        if results and isinstance(results[0], Boto3Model):
            return PrimaryBoto3ModelQuerySet(results)
        return results

    def get(
        self,
        ServiceNamespace: Literal[
            "ecs",
            "elasticmapreduce",
            "ec2",
            "appstream",
            "dynamodb",
            "rds",
            "sagemaker",
            "custom-resource",
            "comprehend",
            "lambda",
            "cassandra",
            "kafka",
            "elasticache",
            "neptune",
            "workspaces",
        ],
        *,
        ResourceIds: List[str] | None = None,
        ScalableDimension: Literal["ecs:service:DesiredCount", "ec2:spot-fleet-request:TargetCapacity", "elasticmapreduce:instancegroup:InstanceCount", "appstream:fleet:DesiredCapacity", "dynamodb:table:ReadCapacityUnits", "dynamodb:table:WriteCapacityUnits", "dynamodb:index:ReadCapacityUnits", "dynamodb:index:WriteCapacityUnits", "rds:cluster:ReadReplicaCount", "sagemaker:variant:DesiredInstanceCount", "custom-resource:ResourceType:Property", "comprehend:document-classifier-endpoint:DesiredInferenceUnits", "comprehend:entity-recognizer-endpoint:DesiredInferenceUnits", "lambda:function:ProvisionedConcurrency", "cassandra:table:ReadCapacityUnits", "cassandra:table:WriteCapacityUnits", "kafka:broker-storage:VolumeSize", "elasticache:cache-cluster:Nodes", "elasticache:replication-group:NodeGroups", "elasticache:replication-group:Replicas", "neptune:cluster:ReadReplicaCount", "sagemaker:variant:DesiredProvisionedConcurrency", "sagemaker:inference-component:DesiredCopyCount", "workspaces:workspacespool:DesiredUserSessions"] | None = None,
    ) -> Optional["ScalableTarget"]:
        """
        Gets information about the scalable targets in the specified namespace.

        Args:
            ServiceNamespace: The namespace of the Amazon Web Services service that provides the resource. For a resource
                provided by your own application or service, use ``custom-resource`` instead.

        Keyword Args:
            ResourceIds: The identifier of the resource associated with the scalable target. This string consists of the
                resource type and unique identifier.
            ScalableDimension: The scalable dimension associated with the scalable target. This string consists of the service
                namespace, resource type, and scaling property. If you specify a scalable dimension, you must also specify a
                resource ID.
        """
        args: Dict[str, Any] = dict(
            ServiceNamespace=self.serialize(ServiceNamespace),
            ResourceIds=self.serialize(ResourceIds),
            ScalableDimension=self.serialize(ScalableDimension),
        )
        _response = self.client.describe_scalable_targets(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = DescribeScalableTargetsResponse(**_response)

        if response and response.ScalableTargets:
            self.sessionize(response.ScalableTargets[0])
            return response.ScalableTargets[0]
        return None


class ScheduledActionManager(Boto3ModelManager):
    service_name: str = "application-autoscaling"

    def create(self, model: "ScheduledAction") -> None:
        """
        Creates or updates a scheduled action for an Application Auto Scaling scalable
        target.

        Args:
            model: The :py:class:`ScheduledAction` to create.
        """
        data = model.model_dump(exclude_none=True, by_alias=True)
        args = dict(
            ServiceNamespace=data.get("ServiceNamespace"),
            ScheduledActionName=data.get("ScheduledActionName"),
            ResourceId=data.get("ResourceId"),
            ScalableDimension=data.get("ScalableDimension"),
            Schedule=data.get("Schedule"),
            Timezone=data.get("Timezone"),
            StartTime=data.get("StartTime"),
            EndTime=data.get("EndTime"),
            ScalableTargetAction=data.get("ScalableTargetAction"),
        )
        self.client.put_scheduled_action(
            **{k: v for k, v in args.items() if v is not None}
        )

    def update(self, model: "ScheduledAction") -> None:
        """
        Creates or updates a scheduled action for an Application Auto Scaling scalable
        target.

        Args:
            model: The :py:class:`ScheduledAction` to update.
        """
        data = model.model_dump(exclude_none=True, by_alias=True)
        args = dict(
            ServiceNamespace=data.get("ServiceNamespace"),
            ScheduledActionName=data.get("ScheduledActionName"),
            ResourceId=data.get("ResourceId"),
            ScalableDimension=data.get("ScalableDimension"),
            Schedule=data.get("Schedule"),
            Timezone=data.get("Timezone"),
            StartTime=data.get("StartTime"),
            EndTime=data.get("EndTime"),
            ScalableTargetAction=data.get("ScalableTargetAction"),
        )
        self.client.put_scheduled_action(
            **{k: v for k, v in args.items() if v is not None}
        )

    def delete(
        self,
        ServiceNamespace: Literal[
            "ecs",
            "elasticmapreduce",
            "ec2",
            "appstream",
            "dynamodb",
            "rds",
            "sagemaker",
            "custom-resource",
            "comprehend",
            "lambda",
            "cassandra",
            "kafka",
            "elasticache",
            "neptune",
            "workspaces",
        ],
        ScheduledActionName: str,
        ResourceId: str,
        ScalableDimension: Literal[
            "ecs:service:DesiredCount",
            "ec2:spot-fleet-request:TargetCapacity",
            "elasticmapreduce:instancegroup:InstanceCount",
            "appstream:fleet:DesiredCapacity",
            "dynamodb:table:ReadCapacityUnits",
            "dynamodb:table:WriteCapacityUnits",
            "dynamodb:index:ReadCapacityUnits",
            "dynamodb:index:WriteCapacityUnits",
            "rds:cluster:ReadReplicaCount",
            "sagemaker:variant:DesiredInstanceCount",
            "custom-resource:ResourceType:Property",
            "comprehend:document-classifier-endpoint:DesiredInferenceUnits",
            "comprehend:entity-recognizer-endpoint:DesiredInferenceUnits",
            "lambda:function:ProvisionedConcurrency",
            "cassandra:table:ReadCapacityUnits",
            "cassandra:table:WriteCapacityUnits",
            "kafka:broker-storage:VolumeSize",
            "elasticache:cache-cluster:Nodes",
            "elasticache:replication-group:NodeGroups",
            "elasticache:replication-group:Replicas",
            "neptune:cluster:ReadReplicaCount",
            "sagemaker:variant:DesiredProvisionedConcurrency",
            "sagemaker:inference-component:DesiredCopyCount",
            "workspaces:workspacespool:DesiredUserSessions",
        ],
    ) -> None:
        """
        Deletes the specified scheduled action for an Application Auto Scaling scalable
        target.

        Args:
            ServiceNamespace: The namespace of the Amazon Web Services service that provides the resource. For a resource
                provided by your own application or service, use ``custom-resource`` instead.
            ScheduledActionName: The name of the scheduled action.
            ResourceId: The identifier of the resource associated with the scheduled action. This string consists of the
                resource type and unique identifier.
            ScalableDimension: The scalable dimension. This string consists of the service namespace, resource type, and scaling
                property.
        """
        args: Dict[str, Any] = dict(
            ServiceNamespace=self.serialize(ServiceNamespace),
            ScheduledActionName=self.serialize(ScheduledActionName),
            ResourceId=self.serialize(ResourceId),
            ScalableDimension=self.serialize(ScalableDimension),
        )
        self.client.delete_scheduled_action(
            **{k: v for k, v in args.items() if v is not None}
        )

    def list(
        self,
        ServiceNamespace: Literal[
            "ecs",
            "elasticmapreduce",
            "ec2",
            "appstream",
            "dynamodb",
            "rds",
            "sagemaker",
            "custom-resource",
            "comprehend",
            "lambda",
            "cassandra",
            "kafka",
            "elasticache",
            "neptune",
            "workspaces",
        ],
        *,
        ScheduledActionNames: List[str] | None = None,
        ResourceId: str | None = None,
        ScalableDimension: Literal["ecs:service:DesiredCount", "ec2:spot-fleet-request:TargetCapacity", "elasticmapreduce:instancegroup:InstanceCount", "appstream:fleet:DesiredCapacity", "dynamodb:table:ReadCapacityUnits", "dynamodb:table:WriteCapacityUnits", "dynamodb:index:ReadCapacityUnits", "dynamodb:index:WriteCapacityUnits", "rds:cluster:ReadReplicaCount", "sagemaker:variant:DesiredInstanceCount", "custom-resource:ResourceType:Property", "comprehend:document-classifier-endpoint:DesiredInferenceUnits", "comprehend:entity-recognizer-endpoint:DesiredInferenceUnits", "lambda:function:ProvisionedConcurrency", "cassandra:table:ReadCapacityUnits", "cassandra:table:WriteCapacityUnits", "kafka:broker-storage:VolumeSize", "elasticache:cache-cluster:Nodes", "elasticache:replication-group:NodeGroups", "elasticache:replication-group:Replicas", "neptune:cluster:ReadReplicaCount", "sagemaker:variant:DesiredProvisionedConcurrency", "sagemaker:inference-component:DesiredCopyCount", "workspaces:workspacespool:DesiredUserSessions"] | None = None,
    ) -> PrimaryBoto3ModelQuerySet:
        """
        Describes the Application Auto Scaling scheduled actions for the specified
        service namespace.

        Args:
            ServiceNamespace: The namespace of the Amazon Web Services service that provides the resource. For a resource
                provided by your own application or service, use ``custom-resource`` instead.

        Keyword Args:
            ScheduledActionNames: The names of the scheduled actions to describe.
            ResourceId: The identifier of the resource associated with the scheduled action. This string consists of the
                resource type and unique identifier.
            ScalableDimension: The scalable dimension. This string consists of the service namespace, resource type, and scaling
                property. If you specify a scalable dimension, you must also specify a resource ID.
        """
        paginator = self.client.get_paginator("describe_scheduled_actions")
        args: Dict[str, Any] = dict(
            ServiceNamespace=self.serialize(ServiceNamespace),
            ScheduledActionNames=self.serialize(ScheduledActionNames),
            ResourceId=self.serialize(ResourceId),
            ScalableDimension=self.serialize(ScalableDimension),
        )
        response_iterator = paginator.paginate(
            **{k: v for k, v in args.items() if v is not None}
        )
        results = []
        for _response in response_iterator:
            if list(_response.keys()) == ["ResponseMetadata"]:
                break
            if "ResponseMetadata" in _response:
                del _response["ResponseMetadata"]
            response = DescribeScheduledActionsResponse(**_response)
            if response.ScheduledActions:
                results.extend(response.ScheduledActions)
            else:
                if getattr(response, "NextToken", None):
                    continue
                break
        self.sessionize(results)
        if results and isinstance(results[0], Boto3Model):
            return PrimaryBoto3ModelQuerySet(results)
        return results

    def get(
        self,
        ServiceNamespace: Literal[
            "ecs",
            "elasticmapreduce",
            "ec2",
            "appstream",
            "dynamodb",
            "rds",
            "sagemaker",
            "custom-resource",
            "comprehend",
            "lambda",
            "cassandra",
            "kafka",
            "elasticache",
            "neptune",
            "workspaces",
        ],
        *,
        ScheduledActionNames: List[str] | None = None,
        ResourceId: str | None = None,
        ScalableDimension: Literal["ecs:service:DesiredCount", "ec2:spot-fleet-request:TargetCapacity", "elasticmapreduce:instancegroup:InstanceCount", "appstream:fleet:DesiredCapacity", "dynamodb:table:ReadCapacityUnits", "dynamodb:table:WriteCapacityUnits", "dynamodb:index:ReadCapacityUnits", "dynamodb:index:WriteCapacityUnits", "rds:cluster:ReadReplicaCount", "sagemaker:variant:DesiredInstanceCount", "custom-resource:ResourceType:Property", "comprehend:document-classifier-endpoint:DesiredInferenceUnits", "comprehend:entity-recognizer-endpoint:DesiredInferenceUnits", "lambda:function:ProvisionedConcurrency", "cassandra:table:ReadCapacityUnits", "cassandra:table:WriteCapacityUnits", "kafka:broker-storage:VolumeSize", "elasticache:cache-cluster:Nodes", "elasticache:replication-group:NodeGroups", "elasticache:replication-group:Replicas", "neptune:cluster:ReadReplicaCount", "sagemaker:variant:DesiredProvisionedConcurrency", "sagemaker:inference-component:DesiredCopyCount", "workspaces:workspacespool:DesiredUserSessions"] | None = None,
    ) -> Optional["ScheduledAction"]:
        """
        Describes the Application Auto Scaling scheduled actions for the specified
        service namespace.

        Args:
            ServiceNamespace: The namespace of the Amazon Web Services service that provides the resource. For a resource
                provided by your own application or service, use ``custom-resource`` instead.

        Keyword Args:
            ScheduledActionNames: The names of the scheduled actions to describe.
            ResourceId: The identifier of the resource associated with the scheduled action. This string consists of the
                resource type and unique identifier.
            ScalableDimension: The scalable dimension. This string consists of the service namespace, resource type, and scaling
                property. If you specify a scalable dimension, you must also specify a resource ID.
        """
        args: Dict[str, Any] = dict(
            ServiceNamespace=self.serialize(ServiceNamespace),
            ScheduledActionNames=self.serialize(ScheduledActionNames),
            ResourceId=self.serialize(ResourceId),
            ScalableDimension=self.serialize(ScalableDimension),
        )
        _response = self.client.describe_scheduled_actions(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = DescribeScheduledActionsResponse(**_response)

        if response and response.ScheduledActions:
            self.sessionize(response.ScheduledActions[0])
            return response.ScheduledActions[0]
        return None


# ==============
# Service Models
# ==============


class StepAdjustment(Boto3Model):
    """Represents a step adjustment for a `StepScalingPolicyConfiguration <https://docs.aws.amazon.com/autoscaling/application/
    APIReference/API_StepScalingPolicyConfiguration.html>`_. Describes an adjustment based on the difference between the value
    of the aggregated CloudWatch metric and the breach threshold that you've defined for the alarm.

    For the following examples, suppose that you have an alarm with a breach threshold of 50:

    * To initiate the adjustment when the metric is greater than or equal to 50 and less than 60, specify a lower bound of
      ``0`` and an upper bound of ``10``.
    * To initiate the adjustment when the metric is greater than 40 and less than or equal to 50, specify a lower bound of
      ``-10`` and an upper bound of ``0``.

    There are a few rules for the step adjustments for your step policy:

    * The ranges of your step adjustments can't overlap or have a gap.
    * At most one step adjustment can have a null lower bound. If one step adjustment has a negative lower bound, then there
      must be a step adjustment with a null lower bound.
    * At most one step adjustment can have a null upper bound. If one step adjustment has a positive upper bound, then there
      must be a step adjustment with a null upper bound.
    * The upper and lower bound can't be null in the same step adjustment.
    """

    MetricIntervalLowerBound: float | None = None
    """
    The lower bound for the difference between the alarm threshold and the CloudWatch
    metric.

    If the metric value is above the breach threshold, the lower bound is inclusive (the
    metric must be greater than or equal to the threshold plus the lower bound).
    Otherwise, it's exclusive (the metric must be greater than the threshold plus the
    lower bound). A null value indicates negative infinity.
    """
    MetricIntervalUpperBound: float | None = None
    """
    The upper bound for the difference between the alarm threshold and the CloudWatch
    metric.

    If the metric value is above the breach threshold, the upper bound is exclusive (the
    metric must be less than the threshold plus the upper bound). Otherwise, it's
    inclusive (the metric must be less than or equal to the threshold plus the upper
    bound). A null value indicates positive infinity.
    """
    ScalingAdjustment: int
    """
    The amount by which to scale, based on the specified adjustment type.

    A positive value adds to the current capacity while a negative number removes from
    the current capacity. For exact capacity, you must specify a non-negative value.
    """


class ApplicationAutoscalingStepScalingPolicyConfiguration(Boto3Model):
    """
    Represents a step scaling policy configuration to use with Application Auto Scaling.

    For more information, see
    `Step scaling policies <https://docs.aws.amazon.com/autoscaling/application/userguide/application-auto-scaling-step-scaling-
    policies.html>`_ in the *Application Auto Scaling User Guide*.
    """

    AdjustmentType: Literal["ChangeInCapacity", "PercentChangeInCapacity", "ExactCapacity"] | None = None
    """
Specifies how the ``ScalingAdjustment`` value in a
`StepAdjustment <https://docs.aws.amazon.com/autoscaling/application/APIReference/API_StepAdjustment.html>`_ is
interpreted (for example, an absolute number or a percentage). The valid values are ``ChangeInCapacity``,
``ExactCapacity``, and ``PercentChangeInCapacity``.
    """
    StepAdjustments: List["StepAdjustment"] | None = None
    """
    A set of adjustments that enable you to scale based on the size of the alarm breach.
    """
    MinAdjustmentMagnitude: int | None = None
    """
    The minimum value to scale by when the adjustment type is
    ``PercentChangeInCapacity``.

    For example, suppose that you
    create a step scaling policy to scale out an Amazon ECS service by 25 percent and you specify a
    ``MinAdjustmentMagnitude`` of 2. If the service has 4 tasks and the scaling policy is performed, 25 percent of 4 is 1.
    However, because you specified a ``MinAdjustmentMagnitude`` of 2, Application Auto Scaling scales out the service by 2
    tasks.
    """
    Cooldown: int | None = None
    """
    The amount of time, in seconds, to wait for a previous scaling activity to take
    effect.

    If not specified, the default value is 300. For more information, see
    `Cooldown period <https://docs.aws.amazon.com/autoscaling/application/userguide/step-scaling-policy-overview.html#step-scaling-
    cooldown>`_ in the *Application Auto Scaling User Guide*.
    """
    MetricAggregationType: Literal["Average", "Minimum", "Maximum"] | None = None
    """
    The aggregation type for the CloudWatch metrics.

    Valid values are ``Minimum``, ``Maximum``, and ``Average``. If the
    aggregation type is null, the value is treated as ``Average``.
    """


class ApplicationAutoscalingPredefinedMetricSpecification(Boto3Model):
    """
    Represents a predefined metric for a target tracking scaling policy to use with
    Application Auto Scaling.

    For more information,
    `Predefined metrics for target tracking scaling policies <https://docs.aws.amazon.com/autoscaling/application/userguide/monitoring-cloudwatch.html#predefined-metrics>`_
    in the *Application Auto Scaling User Guide*.
    """

    PredefinedMetricType: Literal[
        "DynamoDBReadCapacityUtilization",
        "DynamoDBWriteCapacityUtilization",
        "ALBRequestCountPerTarget",
        "RDSReaderAverageCPUUtilization",
        "RDSReaderAverageDatabaseConnections",
        "EC2SpotFleetRequestAverageCPUUtilization",
        "EC2SpotFleetRequestAverageNetworkIn",
        "EC2SpotFleetRequestAverageNetworkOut",
        "SageMakerVariantInvocationsPerInstance",
        "ECSServiceAverageCPUUtilization",
        "ECSServiceAverageMemoryUtilization",
        "AppStreamAverageCapacityUtilization",
        "ComprehendInferenceUtilization",
        "LambdaProvisionedConcurrencyUtilization",
        "CassandraReadCapacityUtilization",
        "CassandraWriteCapacityUtilization",
        "KafkaBrokerStorageUtilization",
        "ElastiCacheEngineCPUUtilization",
        "ElastiCacheDatabaseMemoryUsagePercentage",
        "ElastiCachePrimaryEngineCPUUtilization",
        "ElastiCacheReplicaEngineCPUUtilization",
        "ElastiCacheDatabaseMemoryUsageCountedForEvictPercentage",
        "NeptuneReaderAverageCPUUtilization",
        "SageMakerVariantProvisionedConcurrencyUtilization",
        "ElastiCacheDatabaseCapacityUsageCountedForEvictPercentage",
        "SageMakerInferenceComponentInvocationsPerCopy",
        "WorkSpacesAverageUserSessionsCapacityUtilization",
        "SageMakerInferenceComponentConcurrentRequestsPerCopyHighResolution",
        "SageMakerVariantConcurrentRequestsPerModelHighResolution",
    ]
    """
    The metric type.

    The ``ALBRequestCountPerTarget`` metric type applies only to Spot Fleets and ECS services.
    """
    ResourceLabel: str | None = None
    """
    Identifies the resource associated with the metric type.

    You can't specify a resource label unless the metric type is
    ``ALBRequestCountPerTarget`` and there is a target group attached to the Spot Fleet or ECS service.
    """


class MetricDimension(Boto3Model):
    """
    Describes the dimension names and values associated with a metric.
    """

    Name: str
    """
    The name of the dimension.
    """
    Value: str
    """
    The value of the dimension.
    """


class TargetTrackingMetricDimension(Boto3Model):
    """
    Describes the dimension of a metric.
    """

    Name: str
    """
    The name of the dimension.
    """
    Value: str
    """
    The value of the dimension.
    """


class TargetTrackingMetric(Boto3Model):
    """
    Represents a specific metric.

    Metric is a property of the TargetTrackingMetricStat object.
    """

    Dimensions: List["TargetTrackingMetricDimension"] | None = None
    """
    The dimensions for the metric.

    For the list of available dimensions, see the Amazon Web Services documentation
    available from the table in
    `Amazon Web Services services that publish CloudWatch metrics <https://docs.aws.amazon.com/AmazonCloudWatch/latest/monitoring/aws-services-cloudwatch-metrics.html>`_
    in the
    *Amazon CloudWatch User Guide*.
    """
    MetricName: str | None = None
    """
    The name of the metric.
    """
    Namespace: str | None = None
    """
    The namespace of the metric.

    For more information, see the table in
    `Amazon Web Services services that publish CloudWatch metrics <https://docs.aws.amazon.com/AmazonCloudWatch/latest/monitoring/aws-services-cloudwatch-metrics.html>`_
    in the *Amazon CloudWatch User Guide*.
    """


class TargetTrackingMetricStat(Boto3Model):
    """
    This structure defines the CloudWatch metric to return, along with the statistic and
    unit.

    For more information about the CloudWatch terminology below, see
    `Amazon CloudWatch concepts <https://docs.aws.amazon.com/AmazonCloudWatch/latest/monitoring/cloudwatch_concepts.html>`_
    in the *Amazon
    CloudWatch User Guide*.
    """

    Metric: TargetTrackingMetric
    """
    The CloudWatch metric to return, including the metric name, namespace, and
    dimensions.

    To get the exact metric name, namespace, and dimensions, inspect the
    `Metric <https://docs.aws.amazon.com/AmazonCloudWatch/latest/APIReference/API_Metric.html>`_
    object that is returned by a    call to
    `ListMetrics <https://docs.aws.amazon.com/AmazonCloudWatch/latest/APIReference/API_ListMetrics.html>`_.
    """
    Stat: str
    """
    The statistic to return.

    It can include any CloudWatch statistic or extended statistic. For a list of valid
    values, see the table in
    `Statistics <https://docs.aws.amazon.com/AmazonCloudWatch/latest/monitoring/cloudwatch_concepts.html#Statistic>`_
    in the
    *Amazon CloudWatch User Guide*.
    """
    Unit: str | None = None
    """
    The unit to use for the returned data points.

    For a complete list of the units that CloudWatch supports, see the
    `MetricDatum <https://docs.aws.amazon.com/AmazonCloudWatch/latest/APIReference/API_MetricDatum.html>`_
    data type in the
    *Amazon CloudWatch API Reference*.
    """


class TargetTrackingMetricDataQuery(Boto3Model):
    """
    The metric data to return. Also defines whether this call is returning data for one
    metric only, or whether it is performing a math expression on the values of returned
    metric statistics to create a new time series. A time series is a series of data
    points, each of which is associated with a timestamp.

    For more information and examples, see
    `Create a target tracking scaling policy for Application Auto Scaling using metric math <https://docs.aws.amazon.com/autoscaling/application/userguide/application-auto-scaling-target-tracking-
    metric-math.html>`_ in the *Application Auto Scaling User Guide*.
    """

    Expression: str | None = None
    """
    The math expression to perform on the returned data, if this object is performing a
    math expression.

    This expression can
    use the ``Id`` of the other metrics to refer to those metrics, and can also use the ``Id`` of other expressions to use
    the result of those expressions.
    """
    Id: str
    """
    A short name that identifies the object's results in the response.

    This name must be unique among all
    ``MetricDataQuery`` objects specified for a single scaling policy. If you are performing math expressions on this set of
    data, this name represents that data and can serve as a variable in the mathematical expression. The valid characters
    are letters, numbers, and underscores. The first character must be a lowercase letter.
    """
    Label: str | None = None
    """
    A human-readable label for this metric or expression.

    This is especially useful if this is a math expression, so that you know what the
    value represents.
    """
    MetricStat: TargetTrackingMetricStat | None = None
    """
    Information about the metric data to return.
    """
    ReturnData: bool | None = None
    """
    Indicates whether to return the timestamps and raw data values of this metric.
    """


class ApplicationAutoscalingCustomizedMetricSpecification(Boto3Model):
    """
    Represents a CloudWatch metric of your choosing for a target tracking scaling policy
    to use with Application Auto Scaling.

    For information about the available metrics for a service, see `Amazon Web Services services that publish CloudWatch
    metrics <https://docs.aws.amazon.com/AmazonCloudWatch/latest/monitoring/aws-services-cloudwatch-metrics.html>`_ in the
    *Amazon CloudWatch User Guide*.

    To create your customized metric specification:

    * Add values for each required parameter from CloudWatch. You can use an existing metric, or a new metric that you
      create. To use your own metric, you must first publish the metric to CloudWatch. For more information, see `Publish
      custom metrics <https://docs.aws.amazon.com/AmazonCloudWatch/latest/monitoring/publishingMetrics.html>`_ in the *Amazon
      CloudWatch User Guide*.
    * Choose a metric that changes proportionally with capacity. The value of the metric should increase or decrease in
      inverse proportion to the number of capacity units. That is, the value of the metric should decrease when capacity
      increases, and increase when capacity decreases.

    For more information about the CloudWatch terminology below, see `Amazon CloudWatch
    concepts <https://docs.aws.amazon.com/AmazonCloudWatch/latest/monitoring/cloudwatch_concepts.html>`_ in the *Amazon
    CloudWatch User Guide*.
    """

    MetricName: str | None = None
    """
    The name of the metric.

    To get the exact metric name, namespace, and dimensions, inspect the
    `Metric <https://docs.aws.amazon.com/AmazonCloudWatch/latest/APIReference/API_Metric.html>`_
    object that's returned by a    call to
    `ListMetrics <https://docs.aws.amazon.com/AmazonCloudWatch/latest/APIReference/API_ListMetrics.html>`_.
    """
    Namespace: str | None = None
    """
    The namespace of the metric.
    """
    Dimensions: List["MetricDimension"] | None = None
    """
    The dimensions of the metric.
    """
    Statistic: Literal["Average", "Minimum", "Maximum", "SampleCount", "Sum"] | None = None
    """
    The statistic of the metric.
    """
    Unit: str | None = None
    """
    The unit of the metric.

    For a complete list of the units that CloudWatch supports, see the
    `MetricDatum <https://docs.aws.amazon.com/AmazonCloudWatch/latest/APIReference/API_MetricDatum.html>`_
    data type in the
    *Amazon CloudWatch API Reference*.
    """
    Metrics: List["TargetTrackingMetricDataQuery"] | None = None
    """
    The metrics to include in the target tracking scaling policy, as a metric data
    query.

    This can include both raw metric and metric math expressions.
    """


class ApplicationAutoscalingTargetTrackingScalingPolicyConfiguration(Boto3Model):
    """
    Represents a target tracking scaling policy configuration to use with Application
    Auto Scaling.

    For more information, see
    `Target tracking scaling policies <https://docs.aws.amazon.com/autoscaling/application/userguide/application-auto-scaling-target-tracking.html>`_
    in the *Application Auto Scaling User Guide*.
    """

    TargetValue: float
    """
    The target value for the metric.

    Although this property accepts numbers of type Double, it won't accept values that
    are either too small or too large. Values must be in the range of -2^360 to 2^360.
    The value must be a valid number based on the choice of metric. For example, if the
    metric is CPU utilization, then the target value is a percent value that represents
    how much of the CPU can be used before scaling out.
    """
    PredefinedMetricSpecification: ApplicationAutoscalingPredefinedMetricSpecification | None = None
    """
    A predefined metric.

    You can specify either a predefined metric or a customized metric.
    """
    CustomizedMetricSpecification: ApplicationAutoscalingCustomizedMetricSpecification | None = None
    """
    A customized metric.

    You can specify either a predefined metric or a customized metric.
    """
    ScaleOutCooldown: int | None = None
    """
    The amount of time, in seconds, to wait for a previous scale-out activity to take
    effect.

    For more information and for
    default values, see `Define cooldown periods <https://docs.aws.amazon.com/autoscaling/application/userguide/target-
    tracking-scaling-policy-overview.html#target-tracking-cooldown>`_ in the *Application Auto Scaling User Guide*.
    """
    ScaleInCooldown: int | None = None
    """
    The amount of time, in seconds, after a scale-in activity completes before another
    scale-in activity can start.

    For more information and for default values, see
    `Define cooldown periods <https://docs.aws.amazon.com/autoscaling/application/userguide/target-tracking-scaling-policy-
    overview.html#target-tracking-cooldown>`_ in the *Application Auto Scaling User Guide*.
    """
    DisableScaleIn: bool | None = None
    """
    Indicates whether scale in by the target tracking scaling policy is disabled.

    If the value is ``true``, scale in is
    disabled and the target tracking scaling policy won't remove capacity from the scalable target. Otherwise, scale in is
    enabled and the target tracking scaling policy can remove capacity from the scalable target. The default value is
    ``false``.
    """


class PredictiveScalingPredefinedMetricPairSpecification(Boto3Model):
    """
    Represents a metric pair for a predictive scaling policy.

    The following predefined metrics are available for predictive scaling:

    * ``ECSServiceAverageCPUUtilization``
    * ``ECSServiceAverageMemoryUtilization``
    * ``ECSServiceCPUUtilization``
    * ``ECSServiceMemoryUtilization``
    * ``ECSServiceTotalCPUUtilization``
    * ``ECSServiceTotalMemoryUtilization``
    * ``ALBRequestCount``
    * ``ALBRequestCountPerTarget``
    * ``TotalALBRequestCount``
    """

    PredefinedMetricType: str
    """
    Indicates which metrics to use.

    There are two different types of metrics for each metric type: one is a load metric and
    one is a scaling metric.
    """
    ResourceLabel: str | None = None
    """
    A label that uniquely identifies a specific target group from which to determine the
    total and average request count.
    """


class PredictiveScalingPredefinedScalingMetricSpecification(Boto3Model):
    """
    Describes a scaling metric for a predictive scaling policy.

    When returned in the output of ``DescribePolicies``, it indicates that a predictive scaling policy uses individually
    specified load and scaling metrics instead of a metric pair.

    The following predefined metrics are available for predictive scaling:

    * ``ECSServiceAverageCPUUtilization``
    * ``ECSServiceAverageMemoryUtilization``
    * ``ECSServiceCPUUtilization``
    * ``ECSServiceMemoryUtilization``
    * ``ECSServiceTotalCPUUtilization``
    * ``ECSServiceTotalMemoryUtilization``
    * ``ALBRequestCount``
    * ``ALBRequestCountPerTarget``
    * ``TotalALBRequestCount``
    """

    PredefinedMetricType: str
    """
    The metric type.
    """
    ResourceLabel: str | None = None
    """
    A label that uniquely identifies a specific target group from which to determine the
    average request count.
    """


class PredictiveScalingPredefinedLoadMetricSpecification(Boto3Model):
    """
    Describes a load metric for a predictive scaling policy.

    When returned in the output of ``DescribePolicies``, it indicates that a predictive scaling policy uses individually
    specified load and scaling metrics instead of a metric pair.

    The following predefined metrics are available for predictive scaling:

    * ``ECSServiceAverageCPUUtilization``
    * ``ECSServiceAverageMemoryUtilization``
    * ``ECSServiceCPUUtilization``
    * ``ECSServiceMemoryUtilization``
    * ``ECSServiceTotalCPUUtilization``
    * ``ECSServiceTotalMemoryUtilization``
    * ``ALBRequestCount``
    * ``ALBRequestCountPerTarget``
    * ``TotalALBRequestCount``
    """

    PredefinedMetricType: str
    """
    The metric type.
    """
    ResourceLabel: str | None = None
    """
    A label that uniquely identifies a target group.
    """


class PredictiveScalingMetricDimension(Boto3Model):
    """
    Describes the dimension of a metric.
    """

    Name: str
    """
    The name of the dimension.
    """
    Value: str
    """
    The value of the dimension.
    """


class PredictiveScalingMetric(Boto3Model):
    """
    Describes the scaling metric.
    """

    Dimensions: List["PredictiveScalingMetricDimension"] | None = None
    """
    Describes the dimensions of the metric.
    """
    MetricName: str | None = None
    """
    The name of the metric.
    """
    Namespace: str | None = None
    """
    The namespace of the metric.
    """


class PredictiveScalingMetricStat(Boto3Model):
    """
    This structure defines the CloudWatch metric to return, along with the statistic and
    unit.
    """

    Metric: PredictiveScalingMetric
    """
    The CloudWatch metric to return, including the metric name, namespace, and
    dimensions.

    To get the exact metric name, namespace, and dimensions, inspect the
    `Metric <https://docs.aws.amazon.com/AmazonCloudWatch/latest/APIReference/API_Metric.html>`_
    object that is returned by a    call to
    `ListMetrics <https://docs.aws.amazon.com/AmazonCloudWatch/latest/APIReference/API_ListMetrics.html>`_.
    """
    Stat: str
    """
    The statistic to return.

    It can include any CloudWatch statistic or extended statistic. For a list of valid
    values, see the table in
    `Statistics <https://docs.aws.amazon.com/AmazonCloudWatch/latest/monitoring/cloudwatch_concepts.html#Statistic>`_
    in the
    *Amazon CloudWatch User Guide*.
    """
    Unit: str | None = None
    """
    The unit to use for the returned data points.

    For a complete list of the units that CloudWatch supports, see the
    `MetricDatum <https://docs.aws.amazon.com/AmazonCloudWatch/latest/APIReference/API_MetricDatum.html>`_
    data type in the
    *Amazon CloudWatch API Reference*.
    """


class PredictiveScalingMetricDataQuery(Boto3Model):
    """
    The metric data to return.

    Also defines whether this call is returning data for one metric only, or whether it
    is performing a math expression on the values of returned metric statistics to
    create a new time series. A time series is a series of data points, each of which is
    associated with a timestamp.
    """

    Id: str
    """
    A short name that identifies the object's results in the response.

    This name must be unique among all
    ``MetricDataQuery`` objects specified for a single scaling policy. If you are performing math expressions on this set of
    data, this name represents that data and can serve as a variable in the mathematical expression. The valid characters
    are letters, numbers, and underscores. The first character must be a lowercase letter.
    """
    Expression: str | None = None
    """
    The math expression to perform on the returned data, if this object is performing a
    math expression.

    This expression can
    use the ``Id`` of the other metrics to refer to those metrics, and can also use the ``Id`` of other expressions to use
    the result of those expressions.
    """
    MetricStat: PredictiveScalingMetricStat | None = None
    """
    Information about the metric data to return.
    """
    Label: str | None = None
    """
    A human-readable label for this metric or expression.

    This is especially useful if this is a math expression, so that you know what the
    value represents.
    """
    ReturnData: bool | None = None
    """
    Indicates whether to return the timestamps and raw data values of this metric.
    """


class PredictiveScalingCustomizedMetricSpecification(Boto3Model):
    """
    Represents a CloudWatch metric of your choosing for a predictive scaling policy.
    """

    MetricDataQueries: List["PredictiveScalingMetricDataQuery"]
    """
    One or more metric data queries to provide data points for a metric specification.
    """


class PredictiveScalingMetricSpecification(Boto3Model):
    """
    This structure specifies the metrics and target utilization settings for a
    predictive scaling policy.

    You must specify either a metric pair, or a load metric and a scaling metric
    individually. Specifying a metric pair instead of individual metrics provides a
    simpler way to configure metrics for a scaling policy. You choose the metric pair,
    and the policy automatically knows the correct sum and average statistics to use for
    the load metric and the scaling metric.
    """

    TargetValue: float
    """
    Specifies the target utilization.
    """
    PredefinedMetricPairSpecification: PredictiveScalingPredefinedMetricPairSpecification | None = None
    """
    The predefined metric pair specification that determines the appropriate scaling
    metric and load metric to use.
    """
    PredefinedScalingMetricSpecification: PredictiveScalingPredefinedScalingMetricSpecification | None = None
    """
    The predefined scaling metric specification.
    """
    PredefinedLoadMetricSpecification: PredictiveScalingPredefinedLoadMetricSpecification | None = None
    """
    The predefined load metric specification.
    """
    CustomizedScalingMetricSpecification: PredictiveScalingCustomizedMetricSpecification | None = None
    """
    The customized scaling metric specification.
    """
    CustomizedLoadMetricSpecification: PredictiveScalingCustomizedMetricSpecification | None = None
    """
    The customized load metric specification.
    """
    CustomizedCapacityMetricSpecification: PredictiveScalingCustomizedMetricSpecification | None = None
    """
    The customized capacity metric specification.
    """


class ApplicationAutoscalingPredictiveScalingPolicyConfiguration(Boto3Model):
    """
    Represents a predictive scaling policy configuration.

    Predictive scaling is supported on Amazon ECS services.
    """

    MetricSpecifications: List["PredictiveScalingMetricSpecification"]
    """
    This structure includes the metrics and target utilization to use for predictive
    scaling.
    """
    Mode: Literal["ForecastOnly", "ForecastAndScale"] | None = None
    """
    The predictive scaling mode.

    Defaults to ``ForecastOnly`` if not specified.
    """
    SchedulingBufferTime: int | None = None
    """
    The amount of time, in seconds, that the start time can be advanced.
    """
    MaxCapacityBreachBehavior: Literal["HonorMaxCapacity", "IncreaseMaxCapacity"] | None = None
    """
    Defines the behavior that should be applied if the forecast capacity approaches or
    exceeds the maximum capacity.

    Defaults to ``HonorMaxCapacity`` if not specified.
    """
    MaxCapacityBuffer: int | None = None
    """
    The size of the capacity buffer to use when the forecast capacity is close to or
    exceeds the maximum capacity.

    The value is specified as a percentage relative to the forecast capacity. For
    example, if the buffer is 10, this means a 10 percent buffer, such that if the
    forecast capacity is 50, and the maximum capacity is 40, then the effective maximum
    capacity is 55.
    """


class Alarm(Boto3Model):
    """
    Represents a CloudWatch alarm associated with a scaling policy.
    """

    AlarmName: str
    """
    The name of the alarm.
    """
    AlarmARN: str
    """
    The Amazon Resource Name (ARN) of the alarm.
    """


class ScalingPolicy(PrimaryBoto3Model):
    """
    Represents a scaling policy to use with Application Auto Scaling.

    For more information about configuring scaling policies for a specific service, see
    `Amazon Web Services services that you can use with Application Auto Scaling <https://docs.aws.amazon.com/autoscaling/application/userguide/integrated-
    services-list.html>`_ in the *Application Auto Scaling User Guide*.
    """

    manager_class: ClassVar[Type[Boto3ModelManager]] = ScalingPolicyManager

    PolicyARN: str
    """
    The Amazon Resource Name (ARN) of the scaling policy.
    """
    PolicyName: str
    """
    The name of the scaling policy.
    """
    ServiceNamespace: Literal[
        "ecs",
        "elasticmapreduce",
        "ec2",
        "appstream",
        "dynamodb",
        "rds",
        "sagemaker",
        "custom-resource",
        "comprehend",
        "lambda",
        "cassandra",
        "kafka",
        "elasticache",
        "neptune",
        "workspaces",
    ]
    """
    The namespace of the Amazon Web Services service that provides the resource, or a
    ``custom-resource``.
    """
    ResourceId: str
    """
    The identifier of the resource associated with the scaling policy.

    This string consists of the resource type and unique identifier.
    """
    ScalableDimension: Literal[
        "ecs:service:DesiredCount",
        "ec2:spot-fleet-request:TargetCapacity",
        "elasticmapreduce:instancegroup:InstanceCount",
        "appstream:fleet:DesiredCapacity",
        "dynamodb:table:ReadCapacityUnits",
        "dynamodb:table:WriteCapacityUnits",
        "dynamodb:index:ReadCapacityUnits",
        "dynamodb:index:WriteCapacityUnits",
        "rds:cluster:ReadReplicaCount",
        "sagemaker:variant:DesiredInstanceCount",
        "custom-resource:ResourceType:Property",
        "comprehend:document-classifier-endpoint:DesiredInferenceUnits",
        "comprehend:entity-recognizer-endpoint:DesiredInferenceUnits",
        "lambda:function:ProvisionedConcurrency",
        "cassandra:table:ReadCapacityUnits",
        "cassandra:table:WriteCapacityUnits",
        "kafka:broker-storage:VolumeSize",
        "elasticache:cache-cluster:Nodes",
        "elasticache:replication-group:NodeGroups",
        "elasticache:replication-group:Replicas",
        "neptune:cluster:ReadReplicaCount",
        "sagemaker:variant:DesiredProvisionedConcurrency",
        "sagemaker:inference-component:DesiredCopyCount",
        "workspaces:workspacespool:DesiredUserSessions",
    ]
    """
    The scalable dimension.

    This string consists of the service namespace, resource type, and scaling property.
    """
    PolicyType: Literal["StepScaling", "TargetTrackingScaling", "PredictiveScaling"]
    """
    The scaling policy type.
    """
    StepScalingPolicyConfiguration: ApplicationAutoscalingStepScalingPolicyConfiguration | None = None
    """
    A step scaling policy.
    """
    TargetTrackingScalingPolicyConfiguration: ApplicationAutoscalingTargetTrackingScalingPolicyConfiguration | None = None
    """
    A target tracking scaling policy.
    """
    PredictiveScalingPolicyConfiguration: ApplicationAutoscalingPredictiveScalingPolicyConfiguration | None = None
    """
    The predictive scaling policy configuration.
    """
    Alarms: List["Alarm"] | None = None
    """
    The CloudWatch alarms associated with the scaling policy.
    """
    CreationTime: datetime
    """
    The Unix timestamp for when the scaling policy was created.
    """

    @property
    def arn(self) -> str | None:
        """
        Return the ARN of the model.   This is the value of the :py:attr:`PolicyARN`
        attribute.

        Returns:
            The ARN of the model instance.
        """
        return self.PolicyARN

    @property
    def name(self) -> str | None:
        """
        Return the name of the model.   This is the value of the :py:attr:`PolicyName`
        attribute.

        Returns:
            The name of the model instance.
        """
        return self.PolicyName

    @property
    def pk(self) -> OrderedDict[str, Any]:
        """
        The primary key of the scaling policy.

        This can be used to get the scaling policy with the :py:meth:`botocraft.services.application_autoscaling.ScalingPolicyManager.get` method, or to do :py:meth:`botocraft.services.application_autoscaling.ScalingPolicyManager.delete`.
        """

        return OrderedDict(
            {
                "PolicyName": self.PolicyName,
                "ServiceNamespace": self.ServiceNamespace,
            }
        )

    @cached_property
    def scalable_target(self) -> Optional["ScalableTarget"]:
        """
        Return the :py:class:`ScalableTarget` object that this scaling policy applies
        to.

        .. note::

            The output of this property is cached on the model instance, so
            calling this multiple times will not result in multiple calls to the
            AWS API.   If you need a fresh copy of the data, you can re-get the
            model instance from the manager.
        """

        try:
            pk = OrderedDict(
                {
                    "ResourceId": self.ResourceId,
                    "ScalableDimension": self.ScalableDimension,
                    "ServiceNamespace": self.ServiceNamespace,
                }
            )
        except AttributeError:
            return None
        return ScalableTarget.objects.using(self.session).get(**pk)  # type: ignore[arg-type]


class ApplicationAutoscalingSuspendedState(Boto3Model):
    """
    Specifies whether the scaling activities for a scalable target are in a suspended
    state.
    """

    DynamicScalingInSuspended: bool | None = None
    """
    Whether scale in by a target tracking scaling policy or a step scaling policy is
    suspended.

    Set the value to ``true`` if
    you don't want Application Auto Scaling to remove capacity when a scaling policy is triggered. The default is ``false``.
    """
    DynamicScalingOutSuspended: bool | None = None
    """
    Whether scale out by a target tracking scaling policy or a step scaling policy is
    suspended.

    Set the value to ``true``
    if you don't want Application Auto Scaling to add capacity when a scaling policy is triggered. The default is ``false``.
    """
    ScheduledScalingSuspended: bool | None = None
    """
    Whether scheduled scaling is suspended.

    Set the value to ``true`` if you don't want Application Auto Scaling to add or
    remove capacity by initiating scheduled actions. The default is ``false``.
    """


class ScalableTarget(ScalableTargetModelMixin, PrimaryBoto3Model):
    """
    Represents a scalable target.
    """

    manager_class: ClassVar[Type[Boto3ModelManager]] = ScalableTargetManager

    ServiceNamespace: Literal[
        "ecs",
        "elasticmapreduce",
        "ec2",
        "appstream",
        "dynamodb",
        "rds",
        "sagemaker",
        "custom-resource",
        "comprehend",
        "lambda",
        "cassandra",
        "kafka",
        "elasticache",
        "neptune",
        "workspaces",
    ]
    """
    The namespace of the Amazon Web Services service that provides the resource, or a
    ``custom-resource``.
    """
    ResourceId: str
    """
    The identifier of the resource associated with the scalable target.

    This string consists of the resource type and unique identifier.
    """
    ScalableDimension: Literal[
        "ecs:service:DesiredCount",
        "ec2:spot-fleet-request:TargetCapacity",
        "elasticmapreduce:instancegroup:InstanceCount",
        "appstream:fleet:DesiredCapacity",
        "dynamodb:table:ReadCapacityUnits",
        "dynamodb:table:WriteCapacityUnits",
        "dynamodb:index:ReadCapacityUnits",
        "dynamodb:index:WriteCapacityUnits",
        "rds:cluster:ReadReplicaCount",
        "sagemaker:variant:DesiredInstanceCount",
        "custom-resource:ResourceType:Property",
        "comprehend:document-classifier-endpoint:DesiredInferenceUnits",
        "comprehend:entity-recognizer-endpoint:DesiredInferenceUnits",
        "lambda:function:ProvisionedConcurrency",
        "cassandra:table:ReadCapacityUnits",
        "cassandra:table:WriteCapacityUnits",
        "kafka:broker-storage:VolumeSize",
        "elasticache:cache-cluster:Nodes",
        "elasticache:replication-group:NodeGroups",
        "elasticache:replication-group:Replicas",
        "neptune:cluster:ReadReplicaCount",
        "sagemaker:variant:DesiredProvisionedConcurrency",
        "sagemaker:inference-component:DesiredCopyCount",
        "workspaces:workspacespool:DesiredUserSessions",
    ]
    """
    The scalable dimension associated with the scalable target.

    This string consists of the service namespace, resource type, and scaling property.
    """
    MinCapacity: int
    """
    The minimum value to scale to in response to a scale-in activity.
    """
    MaxCapacity: int
    """
    The maximum value to scale to in response to a scale-out activity.
    """
    PredictedCapacity: int = Field(default=None, frozen=True)
    """
    The predicted capacity of the scalable target.
    """
    RoleARN: str
    """
    The ARN of an IAM role that allows Application Auto Scaling to modify the scalable
    target on your behalf.
    """
    CreationTime: datetime = Field(frozen=True)
    """
    The Unix timestamp for when the scalable target was created.
    """
    SuspendedState: ApplicationAutoscalingSuspendedState | None = None
    """
    Specifies whether the scaling activities for a scalable target are in a suspended
    state.
    """
    ScalableTargetARN: str = Field(default=None, frozen=True)
    """
    The ARN of the scalable target.
    """

    @property
    def arn(self) -> str | None:
        """
        Return the ARN of the model.   This is the value of the
        :py:attr:`ScalableTargetARN` attribute.

        Returns:
            The ARN of the model instance.
        """
        return self.ScalableTargetARN

    @property
    def name(self) -> str | None:
        """
        Return the name of the model.   This is the value of the :py:attr:`ResourceId`
        attribute.

        Returns:
            The name of the model instance.
        """
        return self.ResourceId

    @property
    def pk(self) -> OrderedDict[str, Any]:
        """
        The primary key of the scalable target.

        This can be used do :py:meth:`botocraft.services.application_autoscaling.ScalableTargetManager.delete`.
        """

        return OrderedDict(
            {
                "ServiceNamespace": self.ServiceNamespace,
                "ResourceIds": [self.ResourceId],
                "ScalableDimension": self.ScalableDimension,
            }
        )

    @cached_property
    def scaling_policies(self) -> List["ScalingPolicy"] | None:
        """
        Return the :py:class:`ScalingPolicy` objects that this scalable target has.

        .. note::

            The output of this property is cached on the model instance, so
            calling this multiple times will not result in multiple calls to the
            AWS API.   If you need a fresh copy of the data, you can re-get the
            model instance from the manager.
        """

        try:
            pk = OrderedDict(
                {
                    "ServiceNamespace": self.ServiceNamespace,
                    "ResourceIds": [self.ResourceId],
                    "ScalableDimension": self.ScalableDimension,
                }
            )
        except AttributeError:
            return []
        return ScalingPolicy.objects.using(self.session).list(**pk)  # type: ignore[arg-type]

    @cached_property
    def scheduled_action(self) -> Optional["ScheduledAction"]:
        """
        Return the :py:class:`ScheduledAction` objects that this scalable target has.

        .. note::

            The output of this property is cached on the model instance, so
            calling this multiple times will not result in multiple calls to the
            AWS API.   If you need a fresh copy of the data, you can re-get the
            model instance from the manager.
        """

        try:
            pk = OrderedDict(
                {
                    "ServiceNamespace": self.ServiceNamespace,
                    "ResourceId": self.ResourceId,
                    "ScalableDimension": self.ScalableDimension,
                }
            )
        except AttributeError:
            return None
        return ScheduledAction.objects.using(self.session).get(**pk)  # type: ignore[arg-type]


class ApplicationAutoscalingScalableTargetAction(Boto3Model):
    """
    Represents the minimum and maximum capacity for a scheduled action.
    """

    MinCapacity: int | None = None
    """
    The minimum capacity.
    """
    MaxCapacity: int | None = None
    """
    The maximum capacity.
    """


class ScheduledAction(PrimaryBoto3Model):
    """
    Represents a scheduled action.
    """

    manager_class: ClassVar[Type[Boto3ModelManager]] = ScheduledActionManager

    ScheduledActionName: str
    """
    The name of the scheduled action.
    """
    ScheduledActionARN: str
    """
    The Amazon Resource Name (ARN) of the scheduled action.
    """
    ServiceNamespace: Literal[
        "ecs",
        "elasticmapreduce",
        "ec2",
        "appstream",
        "dynamodb",
        "rds",
        "sagemaker",
        "custom-resource",
        "comprehend",
        "lambda",
        "cassandra",
        "kafka",
        "elasticache",
        "neptune",
        "workspaces",
    ]
    """
    The namespace of the Amazon Web Services service that provides the resource, or a
    ``custom-resource``.
    """
    Schedule: str
    """
    The schedule for this action.

    The following formats are supported:
    """
    Timezone: str | None = None
    """
    The time zone used when referring to the date and time of a scheduled action, when
    the scheduled action uses an at or cron expression.
    """
    ResourceId: str
    """
    The identifier of the resource associated with the scaling policy.

    This string consists of the resource type and unique identifier.
    """
    ScalableDimension: Literal["ecs:service:DesiredCount", "ec2:spot-fleet-request:TargetCapacity", "elasticmapreduce:instancegroup:InstanceCount", "appstream:fleet:DesiredCapacity", "dynamodb:table:ReadCapacityUnits", "dynamodb:table:WriteCapacityUnits", "dynamodb:index:ReadCapacityUnits", "dynamodb:index:WriteCapacityUnits", "rds:cluster:ReadReplicaCount", "sagemaker:variant:DesiredInstanceCount", "custom-resource:ResourceType:Property", "comprehend:document-classifier-endpoint:DesiredInferenceUnits", "comprehend:entity-recognizer-endpoint:DesiredInferenceUnits", "lambda:function:ProvisionedConcurrency", "cassandra:table:ReadCapacityUnits", "cassandra:table:WriteCapacityUnits", "kafka:broker-storage:VolumeSize", "elasticache:cache-cluster:Nodes", "elasticache:replication-group:NodeGroups", "elasticache:replication-group:Replicas", "neptune:cluster:ReadReplicaCount", "sagemaker:variant:DesiredProvisionedConcurrency", "sagemaker:inference-component:DesiredCopyCount", "workspaces:workspacespool:DesiredUserSessions"] | None = None
    """
    The scalable dimension.

    This string consists of the service namespace, resource type, and scaling property.
    """
    StartTime: datetime | None = None
    """
    The date and time that the action is scheduled to begin, in UTC.
    """
    EndTime: datetime | None = None
    """
    The date and time that the action is scheduled to end, in UTC.
    """
    ScalableTargetAction: ApplicationAutoscalingScalableTargetAction | None = None
    """
    The new minimum and maximum capacity.

    You can set both values or just one. At the scheduled time, if the current capacity
    is below the minimum capacity, Application Auto Scaling scales out to the minimum
    capacity. If the current capacity is above the maximum capacity, Application Auto
    Scaling scales in to the maximum capacity.
    """
    CreationTime: datetime
    """
    The date and time that the scheduled action was created.
    """

    @property
    def arn(self) -> str | None:
        """
        Return the ARN of the model.   This is the value of the
        :py:attr:`ScheduledActionARN` attribute.

        Returns:
            The ARN of the model instance.
        """
        return self.ScheduledActionARN

    @property
    def name(self) -> str | None:
        """
        Return the name of the model.   This is the value of the
        :py:attr:`ScheduledActionName` attribute.

        Returns:
            The name of the model instance.
        """
        return self.ScheduledActionName

    @property
    def pk(self) -> OrderedDict[str, Any]:
        """
        The primary key of the scheduled action.

        This can be used to get the scheduled action with the :py:meth:`botocraft.services.application_autoscaling.ScheduledActionManager.get` method, or to do :py:meth:`botocraft.services.application_autoscaling.ScheduledActionManager.delete`.
        """

        return OrderedDict(
            {
                "ServiceNamespace": self.ServiceNamespace,
                "ScheduledActionNames": [self.ScheduledActionName],
            }
        )

    @cached_property
    def scalable_target(self) -> Optional["ScalableTarget"]:
        """
        Return the :py:class:`ScalableTarget` object that this scheduled action applies
        to.

        .. note::

            The output of this property is cached on the model instance, so
            calling this multiple times will not result in multiple calls to the
            AWS API.   If you need a fresh copy of the data, you can re-get the
            model instance from the manager.
        """

        try:
            pk = OrderedDict(
                {
                    "ServiceNamespace": self.ServiceNamespace,
                    "ResourceIds": [self.ResourceId],
                    "ScalableDimension": self.ScalableDimension,
                }
            )
        except AttributeError:
            return None
        return ScalableTarget.objects.using(self.session).get(**pk)  # type: ignore[arg-type]


# =======================
# Request/Response Models
# =======================


class PutScalingPolicyResponse(Boto3Model):
    PolicyARN: str
    """
    The Amazon Resource Name (ARN) of the resulting scaling policy.
    """
    Alarms: List["Alarm"] | None = None
    """
    The CloudWatch alarms created for the target tracking scaling policy.
    """


class DeleteScalingPolicyResponse(Boto3Model):
    pass


class DescribeScalingPoliciesResponse(Boto3Model):
    ScalingPolicies: List["ScalingPolicy"] | None = None
    """
    Information about the scaling policies.
    """
    NextToken: str | None = None
    """
    The token required to get the next set of results.

    This value is ``null`` if there are no more results to return.
    """


class RegisterScalableTargetResponse(Boto3Model):
    ScalableTargetARN: str | None = None
    """
    The ARN of the scalable target.
    """


class DeregisterScalableTargetResponse(Boto3Model):
    pass


class DescribeScalableTargetsResponse(Boto3Model):
    ScalableTargets: List["ScalableTarget"] | None = None
    """
    The scalable targets that match the request parameters.
    """
    NextToken: str | None = None
    """
    The token required to get the next set of results.

    This value is ``null`` if there are no more results to return.
    """


class DeleteScheduledActionResponse(Boto3Model):
    pass


class DescribeScheduledActionsResponse(Boto3Model):
    ScheduledActions: List["ScheduledAction"] | None = None
    """
    Information about the scheduled actions.
    """
    NextToken: str | None = None
    """
    The token required to get the next set of results.

    This value is ``null`` if there are no more results to return.
    """



from collections import OrderedDict
from typing import TYPE_CHECKING, Dict, List, cast

from botocraft.sync.models import MethodArgumentDefinition, MethodDocstringDefinition

if TYPE_CHECKING:
    from botocraft.sync.models import ModelManagerMethodArgDefinition
    from botocraft.sync.service import DocumentationFormatter, ServiceGenerator


class ModelManagerMethodGenerator:
    """
    The base class for all model manager methods.  These all methods
    that transparently interact with the manager class on the object's
    behalf so that you don't have to do things like.

    .. code-block:: python

        image = Image.objects.get(pk="my-image")
        scan_results = Image.objects.scan_results(imageIds=[ImageIdentifier(image.imageId)])

    And can instead do:

    .. code-block:: python

        image = Image.objects.get(pk="my-image")
        scan_results = image.scan_results()

    To use this, you subclass it and implement the :py:meth:`body`
    property, which is the body of the method.  You can also override
    any of the following properties to customize the method:

    * :py:meth:`decorator`: The method decorators.
    * :py:meth:`signature`: The method signature.
    * :py:meth:`docstring`: The method docstring.
    * :py:meth:`return_type`: The return type annotation.

    Args:
        generator: The model generator that is creating the model class
        model_name: The name of the model we're generating the property for.

    """  # noqa: E501

    def __init__(
        self,
        generator: "ServiceGenerator",
        model_name: str,
        method_name: str,
    ) -> None:
        #: The generator that is creating the service classes for an AWS Service.
        self.generator = generator
        #: The model generator that is creating the model class
        self.model_generator = self.generator.model_generator
        #: The name of the model we're generating the property for.
        self.model_name = model_name
        #: The definition of the model we're generating the property for.
        self.model_def = self.model_generator.get_model_def(model_name)
        #: The name of the method_name we're generating.
        self.method_name = method_name
        #: The definition of the method we're generating.
        self.method_def = self.model_def.manager_methods[self.method_name]
        #: The definition of the manager whose method we're accessing
        self.manager_def = self.generator.service_def.managers[model_name]

        # Validate that the manager method exists on the manager
        if self.method_def.manager_method not in self.manager_def.methods:
            msg = (
                f"{self.model_name}.{self.method_name}: The manager method "
                f"{self.method_def.manager_method} is not a valid method for the "
                f"manager {model_name}Manager."
            )
            raise ValueError(msg)
        # Validate that the manager method is not one that we've already implemented
        # in our base class
        if self.method_def.manager_method in ["create", "update"]:
            msg = (
                f"{self.model_name}.{self.method_name}: The model method "
                f"{self.model_name}.save() already implements "
                f'"{self.method_def.manager_method}".'
            )
            raise ValueError(msg)
        if self.method_def.manager_method == "delete":
            msg = (
                f"{self.model_name}.{self.method_name}: The model method "
                f"{self.model_name}.delete() already implements "
                f'"{self.method_def.manager_method}".'
            )
            raise ValueError(msg)
        #: The manager generator for the model
        self.manager_method_def = self.manager_def.methods[
            self.method_def.manager_method
        ]
        #: The botocore service model for the service we're generating.
        self.service_model = generator.manager_generator.service_model  # type: ignore[has-type]
        #: The AWS API method for the operation we're generating.  This is
        #: something like ``DescribeInstances``.
        botocore_name = generator.manager_generator.client._PY_TO_OP_NAME[  # type: ignore[has-type]  # noqa: SLF001
            self.manager_method_def.boto3_name
        ]
        #: The botocore operation model for the operation we're generating.
        self.operation_model = self.service_model.operation_model(botocore_name)
        #: The input shape for the boto3 operation
        self.input_shape = self.operation_model.input_shape
        self.method_generator = self.generator.manager_generator.get_method_generator(
            self.model_name,
            self.method_def.manager_method,
            self.manager_method_def,
        )
        self.docformatter = cast(
            "DocumentationFormatter",
            generator.manager_generator.docformatter,  # type: ignore[has-type]
        )

    @property
    def decorators(self) -> str | None:
        """
        The decorators for the method.  If the method definition has decorators
        defined, add those first.

        If :py:attr:`botocraft.sync.models.ModelManagerMethodDefinition.cached``
        is to ``True``, also add the ``@lru_cache`` decorator.

        Returns:
            The decorators for the method, or ``None`` if there are none.

        """
        decorators: List[str] = []
        if self.method_def.decorators:
            for decorator in self.method_def.decorators:
                decorators.append(f"    @{decorator.name}")
                if decorator.import_path:
                    self.generator.imports.add(
                        f"from {decorator.import_path} import {decorator.name}"
                    )
        if self.method_def.cached:
            self.generator.imports.add("from functools import lru_cache")
            decorators.append("    @lru_cache")
        if decorators:
            return "\n".join(decorators)
        return None

    @property
    def return_type(self) -> str:
        """
        Return the return type annotation for the method.  We should read
        the return type from the manager method definition.
        """
        if self.manager_method_def.return_type is not None:
            return self.manager_method_def.return_type
        if self.method_generator.output_shape is not None:
            output_shape = self.method_generator.output_shape
            response_attr = self.method_generator.response_attr
            _return_shape = None
            if response_attr is not None:
                try:
                    response_attr_shape = output_shape.members[
                        cast("str", response_attr)
                    ]
                except KeyError:
                    response_model_def = self.model_generator.get_model_def(
                        output_shape.name
                    )
                    for field, field_data in response_model_def.fields.items():
                        if field_data.rename == response_attr:
                            response_attr_shape = output_shape.members[
                                cast("str", field)
                            ]
                            break
                    else:
                        raise
                _return_shape = response_attr_shape
            else:
                _return_shape = output_shape
            return self.model_generator.shape_converter.convert(
                _return_shape, quote=True, name_only=True
            )
        return "None"

    def get_arg_docstring(self, arg: str) -> str | None:
        """
        Return the docstring for the given argument.

        Args:
            arg: the name of the argument

        Returns:
            The docstring for the argument.

        """
        _arg_name = arg
        arg_def = MethodArgumentDefinition()
        for arg_name, _arg_def in self.manager_method_def.args.items():
            if arg_name == arg:
                arg_def = _arg_def
                break
            if _arg_def.rename == arg:
                arg_def = _arg_def
                _arg_name = arg_name
                break
        if arg_def.docstring:
            return arg_def.docstring
        if self.input_shape is not None:
            if _arg_name in self.input_shape.members:
                return self.input_shape.members[_arg_name].documentation
        return None

    @property
    def docstrings_def(self) -> MethodDocstringDefinition:
        """
        Build the docstring definition for the method.

        Returns:
            A :py:class:`MethodDocstringDefinition` instance.

        """
        docstrings: MethodDocstringDefinition = MethodDocstringDefinition()
        if self.method_def.docstring:
            docstrings.method = self.method_def.docstring
        elif self.manager_method_def.docstring:
            docstrings.method = self.manager_method_def.docstring
        else:
            docstrings.method = self.operation_model.documentation
        for arg in self.method_def.user_args.values():
            docstrings.args[arg.name] = self.get_arg_docstring(arg.name)
        for kwarg in self.method_def.user_keyword_args:
            docstrings.kwargs[kwarg.name] = self.get_arg_docstring(kwarg.name)
        return docstrings

    @property
    def docstring(self) -> str | None:
        """
        Return the docstring for the method.
        """
        return self.docstrings_def.render(self.docformatter)

    @property
    def signature(self) -> str:
        """
        Return the method signature.

        Returns:
            The method signature.

        """
        sig: str = f"    def {self.method_name}(self"
        # First get the args and kwargs we know are on the manager method
        manager_args = self.method_generator.args("method")
        manager_kwargs = self.method_generator.kwargs("method")
        if self.method_def.user_args:
            # Deal with the user_args.  These are arguments that the user of this
            # method still need to supply, rather than getting the value from the
            # object
            user_args = OrderedDict(sorted(self.method_def.user_args.items()))
            for arg in user_args.values():
                if arg.name not in manager_args:
                    # If the argument name is one of the manager method arguments, then
                    # raise an error
                    msg = (
                        f"{self.model_name}.{self.method_name}: The argument {arg.name}"
                        " is not a valid argument for "
                        f"the manager method {self.method_def.manager_method}."
                    )
                    raise ValueError(msg)
                attr_type: str | None = arg.attr_type
                if arg.attr_type is None:
                    # If the argument type is not specified, then use the type
                    # from the manager method
                    attr_type = manager_args[arg.name]
                sig += f", {arg.name}: {attr_type}"
        if self.method_def.user_keyword_args:
            # Deal with the user_keyword_args.  These are keyword arguments that the
            # user of this method still need to supply, rather than getting the value
            # from the object
            for kwarg in self.method_def.user_keyword_args:
                if kwarg.name not in manager_kwargs:
                    # If the keyword argument name is not in the manager method, then
                    # raise an error
                    msg = (
                        f"{self.model_name}.{self.method_name}: The keyword argument"
                        f" {kwarg.name} is not a valid keyword argument for the "
                        f"manager method {self.method_def.manager_method}."
                    )
                    raise ValueError(msg)
                # Building the argument here is a bit more complicated because
                # we need to handle the default value.
                if not kwarg.default:
                    # A manager kwarg string will look like "type = default", so
                    # we need to split it on the " = " and take the second half
                    # to get the default value
                    default = manager_kwargs[kwarg.name].split(" = ")[1]
                else:
                    # We're overriding the default value
                    default = kwarg.default
                attr_type = kwarg.attr_type
                if attr_type is None:
                    # A manager kwarg string will look like "type = default", so
                    # we need to split it on the " = " and take the first half
                    # to get the type
                    attr_type = manager_kwargs[kwarg.name].split(" = ")[0]
                attr_type_default = f"{attr_type} = {default}"
                sig += f", {kwarg.name}: {attr_type_default}"
        return_type = self.return_type
        if '"' not in return_type:
            return_type = f'"{return_type}"'
        sig += f") -> {return_type}:"
        return sig

    @property
    def args(self) -> str:
        """
        Return the method arguments from the method definition only.  These
        are the ones that the user must supply when calling the method.  The other
        arguments are supplied by the object.

        Returns:
            The method arguments.

        """
        args_dict: Dict[int, "ModelManagerMethodArgDefinition"] = (  # noqa: UP037
            self.method_def.args | self.method_def.user_args
        )
        if not args_dict:
            return ""
        # The keys in args_dict are positions (zero-based), and the values are the
        # argument names.  We need to validate that the positions are sequential,
        # and that 0 is in the args_dict.
        if 0 not in args_dict:
            msg = (
                f"{self.model_name}.{self.method_name}: There is no "
                "argument in the 0th position."
            )
            raise ValueError(msg)
        # Validate that we're not missing any argument posiions
        for i in range(1, len(args_dict)):
            if i not in args_dict:
                msg = (
                    f"{self.model_name}.{self.method_name}: Argument at "
                    f"position {i} is missing."
                )
                raise ValueError(msg)
        args: OrderedDict[int, str] = OrderedDict()
        for i in range(len(args_dict)):  # pylint: disable=consider-using-enumerate
            if i in self.method_def.args:
                if args_dict[i].value is not None:
                    args[i] = cast("str", args_dict[i].value)
                else:
                    value = args_dict[i].name
                    if "self." not in value:
                        value = f"self.{value}"
                    args[i] = value
            else:
                args[i] = args_dict[i].name
        return " ".join([f"{arg}, " for arg in args.values()])

    @property
    def kwargs(self) -> str:
        """
        Return the method keyword arguments from the method definition only.  These
        are the ones that the user must supply when calling the method.  The other
        arguments are supplied by the object.

        Returns:
            The method keyword arguments.

        """
        kwargs: List[str] = []
        if self.method_def.keyword_args:
            for arg in self.method_def.keyword_args:
                if arg.value is not None:
                    kwargs.append(f"{arg.name}={arg.value}")
                else:
                    kwargs.append(f"{arg.name}=self.{arg.name}")
        if self.method_def.user_keyword_args:
            kwargs.extend(
                [f"{arg.name}={arg.name}" for arg in self.method_def.user_keyword_args]
            )
        return ", ".join(kwargs)

    @property
    def body(self) -> str:
        """
        Return the method body.

        Returns:
            The method body.

        """
        args = self.args
        kwargs = self.kwargs
        # Get the model alias, if any
        model_alias = self.model_def.alternate_name
        if not model_alias:
            model_alias = self.model_name
        code = f"""
        return (
            cast("{model_alias}Manager", self.objects)  # type: ignore[attr-defined]
            .using(self.session)
            .{self.method_def.manager_method}(
                {args}
                {kwargs}
            )
        )
"""
        return code  # noqa: RET504

    @property
    def code(self) -> str:
        """
        Return the code for the method.

        Returns:
            The code for the method.

        """
        code = """

"""
        decorators = self.decorators
        if decorators is not None:
            code += f"""
{decorators}
"""
        code += f"""
{self.signature}
{self.docstring}
{self.body}
"""
        return code



from functools import wraps
from typing import TYPE_CHECKING, Callable

if TYPE_CHECKING:
    from botocraft.services import (
        PutScalingPolicyResponse,
        ScalableTarget,
        ScalingPolicy,
        ScheduledAction,
        Service,
    )

# ----------
# Decorators
# ----------


def scaling_policy_only(
    func: Callable[..., "PutScalingPolicyResponse"],
) -> Callable[..., "ScalingPolicy"]:
    """
    Wraps application-autoscaling put_scaling_policy methods to return the
    actual :py:class:`botocraft.services.application_autoscaling.ScalingPolicy`
    object instead of the response object.
    """

    @wraps(func)
    def wrapper(*args, **_) -> "ScalingPolicy":
        self = args[0]
        policy = args[1]
        func(*args, **_)
        return self.get(
            PolicyNames=[policy.PolicyName],
            ServiceNamespace=policy.ServiceNamespace,
        )

    return wrapper


def scalable_target_only(
    func: Callable[..., "str"],
) -> Callable[..., "ScalableTarget"]:
    """
    Wraps application-autoscaling register_scalable_target methods to return the
    actual :py:class:`botocraft.services.application_autoscaling.ScalableTarget`
    object instead of the ARN.
    """

    @wraps(func)
    def wrapper(*args, **_) -> "ScalableTarget":
        self = args[0]
        target = args[1]
        func(*args, **_)
        return self.get(
            ServiceNamespace=target.ServiceNamespace,
            ResourceIds=[target.ResourceId],
            ScalableDimension=target.ScalableDimension,
        )

    return wrapper


def scheduled_action_only(
    func: Callable[..., None],
) -> Callable[..., "ScheduledAction"]:
    """
    Wraps application-autoscaling put_scheduled_action methods to return the
    actual :py:class:`botocraft.services.application_autoscaling.ScalableTarget`
    object instead of the ARN.
    """

    @wraps(func)
    def wrapper(*args, **_) -> "ScheduledAction":
        self = args[0]
        action = args[1]
        func(*args, **_)
        return self.get(
            ServiceNamespace=action.ServiceNamespace,
            ScalableActionnamese=[action.ScalableActionName],
        )

    return wrapper


# ----------
# Mixins
# ----------


class ScalableTargetModelMixin:
    """
    A mixin class that extends
    :py:class:`~botocraft.services.application_autoscaling.ScalableTarget`.
    """

    @property
    def resource(self) -> "Service":
        """
        Return the resource object for this scalable target.

        Currently only the ``ecs`` service namespace is supported, which
        returns an :py:class:`~botocraft.services.ecs.Service` object.
        """
        from botocraft.services import Service

        _, cluster, service = self.ResourceId.split("/")  # type: ignore[attr-defined]
        if self.ServiceNamespace == "ecs":  # type: ignore[attr-defined]
            return Service.objects.get(service, cluster=cluster)
        msg = f"ServiceNamespace {self.ServiceNamespace} is not supported yet."  # type: ignore[attr-defined]
        raise NotImplementedError(msg)



from __future__ import annotations

from datetime import datetime  # noqa: TC003
from typing import Any, Dict, List

from pydantic import BaseModel, Field


class AWSAPICallViaCloudTrailItem(BaseModel):
    """
    Represents a resource involved in an AWS API call via CloudTrail.  Contains
    identifiers for AWS resources that were accessed or modified during the API
    call.
    """

    #: The AWS account ID that owns the resource
    accountId: str
    #: The Amazon Resource Name (ARN) of the resource
    ARN: str


class RequestParametersItem(BaseModel):
    """
    Represents an image identifier in an ECR API request.  Used to specify which
    container image is being referenced in operations like BatchGetImage.
    """

    #: The tag of the container image
    imageTag: str


class SessionIssuer(BaseModel):
    """
    Represents the entity that issued the session in the user identity context.
    Contains information about the IAM entity (user or role) that created the
    session.
    """

    #: The AWS account ID that owns the entity
    accountId: str
    #: The unique identifier of the entity
    principalId: str
    #: The type of the entity (e.g., "Role", "User")
    type: str
    #: The ARN of the entity
    arn: str
    #: The name of the entity
    userName: str


class Attributes(BaseModel):
    """
    Represents attributes of the session used for the API call.  Includes
    security-related information about the session.
    """

    #: Indicates whether multi-factor authentication was used to create the session
    mfaAuthenticated: str
    #: The timestamp when the session was created
    creationDate: datetime


class RequestParameters(BaseModel):
    """
    Represents the parameters of an ECR API request.
    Contains the details of what was requested in the API call.
    """

    #: The media types that the API client accepts
    acceptedMediaTypes: List[str]
    #: The AWS account ID that owns the ECR registry
    registryId: str
    #: The name of the ECR repository being accessed
    repositoryName: str
    #: The list of image identifiers being requested
    imageIds: List[RequestParametersItem]


class SessionContext(BaseModel):
    """
    Represents the session context for a user identity.
    Contains detailed information about the session used to make the API call.
    """

    #: Federation data if the session was created through web identity federation
    webIdFederationData: Dict[str, Any]
    #: Information about the entity that issued the session
    sessionIssuer: SessionIssuer
    #: Session attributes like MFA status and creation time
    attributes: Attributes


class UserIdentity(BaseModel):
    """
    Represents the identity of the user that performed the API call.  Contains
    information about the AWS principal (user, role, or service) that made the
    request.
    """

    #: The context of the session used for the API call
    sessionContext: SessionContext
    #: The access key ID used to sign the request
    accessKeyId: str
    #: The AWS account ID that the user belongs to
    accountId: str
    #: The unique identifier of the principal
    principalId: str
    #: The type of the identity (e.g., "AssumedRole", "IAMUser")
    type: str
    #: The ARN of the principal
    arn: str
    #: The AWS service that made the request, if applicable
    invokedBy: str


class AWSAPICallViaCloudTrail(BaseModel):
    """
    Represents the details of an AWS API call as recorded by CloudTrail.
    Contains comprehensive information about the API request, including who made
    it, what was requested, and the response.
    """

    #: The parameters sent with the API request
    requestParameters: RequestParameters
    #: The identity of the user that made the API call
    userIdentity: UserIdentity
    #: The unique identifier of the event
    eventID: str
    #: The AWS region where the request was made
    awsRegion: str
    #: The version of the CloudTrail event format
    eventVersion: str
    #: The response elements returned by the API
    responseElements: Dict[str, Any]
    #: The IP address where the request originated from
    sourceIPAddress: str
    #: The AWS service that the request was made to (e.g., "ecr.amazonaws.com")
    eventSource: str
    #: The resources involved in the API call
    resources: List[AWSAPICallViaCloudTrailItem]
    #: The user agent of the client that made the request
    userAgent: str
    #: The type of the event (e.g., "AwsApiCall")
    eventType: str
    #: The request ID of the API call
    requestID: str
    #: The timestamp when the API call was made
    eventTime: datetime
    #: The name of the API call (e.g., "BatchGetImage")
    eventName: str


class ECRAWSAPICallViaCloudTrailEvent(BaseModel):
    """
    Represents a complete AWS CloudTrail event as delivered by EventBridge.
    Wraps the CloudTrail API call details with EventBridge metadata.
    """

    #: The detailed information about the API call
    detail: AWSAPICallViaCloudTrail
    #: The human-readable type of the event (e.g., "AWS API Call via CloudTrail")
    detail_type: str = Field(..., alias="detail-type")
    #: The AWS resources involved in the event
    resources: List[str]
    #: The unique identifier of the event
    id: str
    #: The source of the event (e.g., "aws.ecr")
    source: str
    #: The timestamp when the event was generated
    time: datetime
    #: The AWS region where the event occurred
    region: str
    #: The version of the event schema
    version: str
    #: The AWS account ID where the event occurred
    account: str



from __future__ import annotations

from datetime import datetime  # noqa: TC003
from typing import List

from pydantic import BaseModel, Field


class ECSServiceDeploymentStateChange(BaseModel):
    """
    The main body of the ECS Service Deployment State Change event.
    """

    #: The name of the event
    eventName: str
    #: The type of the event
    eventType: str
    #: The service deployment ID.
    #:
    #: .. note::
    #:    This is not the same as the
    #:    py:class:`~botocraft.services.ecs.ServiceDeployment` ARN.  Not sure
    #:    what this is, really.
    deploymentId: str
    #: The time the Service Deployment was created
    updatedAt: datetime
    #: The reason the Service Deployment was started
    reason: str | None = None


class ECSServiceDeploymentStateChangeEvent(BaseModel):
    """
    An ECS Service Deployment State Change Event class.

    .. important::

        There's no schema in the EventBridge schema registry for this event,
        so this is a best effort attempt to create a schema for it manually.
    """

    #: Where the meat of the event is
    detail: ECSServiceDeploymentStateChange
    #: The human readable name of the event
    detail_type: str = Field(..., alias="detail-type")
    #: The resources the event applies to.  This will be a list of ARNs
    resources: List[str]
    #: The ID of the event
    id: str
    #: The source of the event, e.g. "aws.ecs"
    source: str
    #: The region the event was generated in
    region: str
    #: The time the event was generated, in UTC
    time: datetime
    #: The version of the schema for the event
    version: str
    #: The account the event was generated in
    account: str



import re
from collections import OrderedDict
from typing import (
    TYPE_CHECKING,
    Final,
    List,
    Literal,
    cast,
)

import inflect

from botocraft.sync.models import (
    ManagerMethodDefinition,
    MethodArgumentDefinition,
    MethodDocstringDefinition,
)

if TYPE_CHECKING:
    import botocore.model

    from botocraft.sync.service import (
        DocumentationFormatter,
        ManagerGenerator,
        ModelGenerator,
        PythonTypeShapeConverter,
    )


class ManagerMethodGenerator:
    """
    The base class for all manager method generators.  This is used to generate
    the code for a single method on a manager class.

    To use this, you subclass it and implement the :py:meth:`body`
    property, which is the body of the method.  You can also override
    any of the following properties to customize the method:

    * :py:meth:`signature`: The method signature.
    * :py:meth:`args`: The positional arguments with types for the method
    * :py:meth:`kwargs`: The keyword arguments with types for the method
    * :py:meth:`operation_call`: the code for the boto3 client call.
    * :py:meth:`return_type`: The return type annotation.
    * :py:meth:`response_class`: The response class to use for the
        boto3 client call.

    Args:
        generator: The generator that is creating the manager class for
            an object.  It has the information about the service and the models,
            and is where we're collecting all our code and imports.

    """

    #: A list of arguments that various boto3 calls use for pagination.  We
    #: want to hide them automatically sometimes.
    PAGINATOR_ARGS: Final[List[str]] = [
        "nextToken",
        "maxResults",
        "MaxResults",
        "NextToken",
        "Marker",
        "MaxRecords",
        "PageSize",
    ]

    #: The botocraft method we're implementing
    method_name: str

    def __init__(
        self,
        generator: "ManagerGenerator",
        model_name: str,
        method_def: ManagerMethodDefinition,
        method_name: str | None = None,
    ):
        if method_name is not None:
            self.method_name = method_name
        self.inflect = inflect.engine()
        #: The generator that is creating the entire manager file.  It
        #: has the information about the service and the models, and
        #: is where we're collecting all our code and imports.
        self.generator = generator
        #: The boto3 client for the service we're generating.
        self.client = generator.client  # type: ignore[has-type]
        #: Our own botocraft config for this model.
        self.method_def = method_def
        #: The botocore service model for the service we're generating.
        self.service_model = generator.service_model  # type: ignore[has-type]
        #: The boto3 operation name for the method we're generating.  This is
        #: something like ``describe_instances``.
        self.boto3_name = self.method_def.boto3_name
        #: The AWS API method for the operation we're generating.  This is
        #: something like ``DescribeInstances``.
        botocore_name = self.client._PY_TO_OP_NAME[self.boto3_name]  # noqa: SLF001
        #: The botocore operation model for the operation we're generating.
        self.operation_model = self.service_model.operation_model(botocore_name)
        #: The input shape for the boto3 operation
        self.input_shape = self.operation_model.input_shape
        #: The output shape for the boto3 operation
        self.output_shape = self.operation_model.output_shape
        #: Our model generator, which we use to generate any response models we
        #: need.
        self.model_generator = cast("ModelGenerator", generator.model_generator)  # type: ignore[has-type]
        #: The converter we use to convert botocore shapes to python types.
        self.shape_converter = cast(
            "PythonTypeShapeConverter",
            generator.shape_converter,  # type: ignore[has-type]
        )
        #: Any botocraft imports we need to add to the manager.
        self.imports = generator.imports
        #: The name of the model itself
        self.model_name = model_name
        #: The botocraft model definition for the model we're working with.
        self.model_def = self.model_generator.get_model_def(self.model_name)
        #: The botocraft name for the model we're working with.  This is the name
        #: of the model itself, or whatever the botocraft config for the model
        #: specifies.
        self.real_model_name = self.model_def.name
        if self.model_def.alternate_name:
            self.real_model_name = self.model_def.alternate_name
        #: The plural of the name of the model itself
        self.model_name_plural = self.inflect.plural(self.model_name)
        if self.model_def.plural:
            self.model_name_plural = self.model_def.plural
        #: Our documentation formatter
        self.docformatter = cast("DocumentationFormatter", generator.docformatter)  # type: ignore[has-type]

    def get_explicit_args_from_request(
        self,
    ) -> OrderedDict[str, MethodArgumentDefinition]:
        """
        Compare the botocore input shape for the operation with the
        :py:class:`ModelAttributeDefinition` dict for the model we're working
        with, and return a dictionary of the arguments that are not part of the
        model.

        Returns:
            A dictionary of argument names to argument definitions.

        """
        model_fields = self.model_generator.botocore_shape_field_defs(self.model_name)
        args: OrderedDict[str, MethodArgumentDefinition] = OrderedDict()
        if self.input_shape is not None:
            for arg in self.input_shape.members:
                if arg in self.method_def.args:
                    # We're explicitly defining this argument in our botocraft
                    # config, so we don't need to do anything here.
                    continue
                if arg not in model_fields:
                    args[arg] = MethodArgumentDefinition(explicit=True)
                    if arg in self.input_shape.required_members:
                        args[arg].required = True
        return args

    def is_required(
        self, arg_name: str, location: Literal["method", "operation"] = "method"
    ) -> bool:
        """
        Determine if the given argument is required for the method signature.

        Args:
            arg_name: the name of the argument

        Keyword Args:
            location: where these arguments will be used.  If ``'method'``, then
                we obey both the boto3 operation definition and the botocraft
                method definition.  If ``'operation'``, then we only obey the
                boto3 operation definition.

        Returns:
            If this argument is required, return ``True``, otherwise return
            ``False``.

        """
        if self.input_shape is None:
            return False
        mapping = self.method_def.args
        if location == "method":
            return arg_name in self.input_shape.required_members or (
                arg_name in mapping and mapping[arg_name].required
            )
        return arg_name in self.input_shape.required_members

    def serialize(self, arg: str) -> str:
        """
        Figure out how to serialize the given argument.

        Args:
            arg: the name of the method argument

        Returns:
            A string that is either the argument name, or the argument name
            wrapped in a call to ``self.serialize()``.

        """
        mapping = self.method_def.args
        arg_def = mapping.get(arg, MethodArgumentDefinition())
        _arg = arg
        if arg_def.value:
            return arg_def.value
        if arg_def.source_arg:
            _arg = arg_def.source_arg
        if not arg_def.raw_value:
            return f"self.serialize({_arg})"
        return _arg

    def _args(  # noqa: PLR0912
        self,
        kind: Literal["args", "kwargs"],
        location: Literal["method", "operation"] = "method",
    ) -> OrderedDict[str, str]:
        """
        Used to generate the arguments for the botocraft method signature.

        If kind == 'args', then we generate the positional arguments, but if
        kind == 'kwargs', then we generate the keyword arguments.

        Args:
            kind: what kind of method arguments to generate.

        Keyword Args:
            location: where these arguments will be used.  If ``'method'``, then
                we only generate the arguments that are required for the method
                signature.  If ``'operation'``, then we generate all the arguments
                that are required for the boto3 operation call.

        Returns:
            An ordered dictionary of argument names to types.

        """
        args: OrderedDict[str, str] = OrderedDict()
        if self.input_shape is None:
            return args
        mapping = self.method_def.args
        for arg_name, arg_shape in self.input_shape.members.items():
            arg_def = mapping.get(arg_name, MethodArgumentDefinition())
            _arg_name = arg_name
            if location == "method" and arg_def.rename:
                _arg_name = arg_def.rename
            if arg_def.hidden:
                # This is a hidden argument, so we don't want to expose it
                # in the method signature or the boto3 call.
                continue
            if location == "method" and arg_def.value:
                # This argument has a specific value, so we don't want to
                # expose it in the method signature
                continue
            if arg_def.python_type:
                python_type = cast("str", arg_def.python_type)
            else:
                python_type = self.shape_converter.convert(arg_shape, quote=True)
            if kind == "args":
                if self.is_required(arg_name, location=location):
                    args[_arg_name] = python_type
            elif not self.is_required(arg_name, location=location):
                default: str | None = arg_def.default if arg_def.default else "None"
                if default == "None":
                    args[_arg_name] = f"Optional[{python_type}] = None"
                else:
                    args[_arg_name] = f"{python_type} = {default}"
        if location == "method":
            for arg_name, arg_def in self.method_def.extra_args.items():
                assert arg_name not in args, (
                    f"{self.model_name}Manager.{self.method_name}: "
                )
                f"extra_args.{arg_name}: already defined"
                assert arg_def.python_type is not None, (
                    f"{self.model_name}Manager.{self.method_name}: "
                )
                f"extra_args.{arg_name}: python_type is required"
                if kind == "args":
                    if arg_def.required is True:
                        args[arg_name] = arg_def.python_type
                elif arg_def.required is False:
                    default = arg_def.default if arg_def.default else "None"
                    if default == "None":
                        args[arg_name] = f"Optional[{arg_def.python_type}] = None"
                    else:
                        args[arg_name] = f"{arg_def.python_type} = {default}"
                if arg_def.imports:
                    # If the method definition has any imports, we need to add them
                    # to our imports list.
                    for import_path in arg_def.imports:
                        self.imports.add(import_path)
        return args

    @property
    def explicit_args(self) -> OrderedDict[str, str]:
        """
        Return the explicit positional arguments for the given method.   These
        are arguments that are not part of the model, but are required in the
        boto3 operation call.

        Returns:
            A dictionary of argument names to types.

        """
        args: OrderedDict[str, str] = OrderedDict()
        if not self.input_shape:
            return args
        for arg_name in self.method_def.explicit_args:
            if arg_name in self.input_shape.required_members:
                args[arg_name] = self.shape_converter.convert(
                    self.input_shape.members[arg_name], quote=True
                )
        return args

    @property
    def explicit_kwargs(self) -> OrderedDict[str, str]:
        """
        Return the explicit positional arguments for the given method.   These
        are arguments that are not part of the model, but can be supplied in the
        boto3 operation call.

        Returns:
            A dictionary of argument names to types.

        """
        args: OrderedDict[str, str] = OrderedDict()
        if not self.input_shape:
            return args
        for arg_name in self.method_def.explicit_kwargs:
            if arg_name not in self.input_shape.required_members:
                args[arg_name] = self.shape_converter.convert(
                    self.input_shape.members[arg_name], quote=True
                )
                arg_def = self.method_def.args.get(arg_name, MethodArgumentDefinition())
                if arg_def.default in [None, "None"]:
                    args[arg_name] = f"Optional[{args[arg_name]}]"
        return args

    def args(
        self, location: Literal["method", "operation"] = "method"
    ) -> OrderedDict[str, str]:
        """
        Return the keyword arguments for the given method.  The positional
        arguments are the arguments are required.  They will include a type.

        Example:
            ..code-block:: python

                {
                    'arg1': 'str'
                    'arg2': 'List[str]'
                }

        Keyword Args:
            location: where these arguments will be used.  If ``'method'``, then
                we only generate the arguments that are required for the method
                signature.  If ``'operation'``, then we generate all the arguments
                that are required for the boto3 operation call.

        Returns:
            A dictionary of argument names to types.

        """
        return self._args("args", location=location)

    def kwargs(
        self, location: Literal["method", "operation"] = "method"
    ) -> OrderedDict[str, str]:
        """
        Return the keyword arguments for the given method.  These apply to both
        the method signature and to the boto3 operation call.

        Keyword arguments are the arguments that are not required.  They will
        include a type and a default value.

        Example:
            ..code-block:: python

                {
                    'arg1': 'str = "default"',
                    'arg2': 'Optional[str] = None'
                }

        Keyword Args:
            location: where these arguments will be used.  If ``'method'``, then
                we only generate the arguments that are required for the method
                signature.  If ``'operation'``, then we generate all the arguments
                that are required for the boto3 operation call.

        Returns:
            A dictionary of argument names to types/defaults.

        """
        return self._args("kwargs", location=location)

    @property
    def operation_args(self) -> str:
        """
        Return the argument string for the boto3 call.  We use the
        output of :py:meth:`args` and :py:meth:`kwargs` to generate
        this string.

        Example:
            ``name=name, description=description``

        """
        args = self.args(location="operation")
        kwargs = self.kwargs(location="operation")
        arg_str = ", ".join([f"{arg}={self.serialize(arg)}" for arg in args])
        if args and kwargs:
            arg_str += ", "
        arg_str += ", ".join([f"{arg}={self.serialize(arg)}" for arg in kwargs])
        return f"args: Dict[str, Any] = dict({arg_str})"

    @property
    def operation_call(self) -> str:
        """
        A body snippet that does the actual boto3 call.  and assigns the
        response to ``_response``.  It then uses response to instantiate
        :py:meth:`response_class`.

        Example:
            ..code-block:: python

                _response = self.client.create(name=name, description=description)
                response = ResponseModel(**_response)

        Returns:
            The code for the boto3 call.

        """
        call = self.operation_args
        call = f"self.client.{self.boto3_name}(**{{k: v for k, v in args.items() if v is not None}})"  # noqa: E501
        if self.return_type in ("None", '"None"'):
            return call
        call = "_response = " + call
        call += f"\n        response = {self.response_class}(**_response)"
        return call

    @property
    def response_class(self) -> str:
        """
        Create the response class, add it to the list of response classes, and
        return the type string.  The response class is the class that we use to
        deserialize the boto3 response into a pydantic model.

        If there is no output shape defined in botocore for the boto3 operation, then
        we return ``None``.

        Returns:
            The name of the response class.

        """
        if self.output_shape is None:
            return "None"
        model_name = self.output_shape.name
        # See if we have an alternate name for the model
        response_model_def = self.model_generator.get_model_def(model_name)
        if response_model_def.alternate_name:
            model_name = response_model_def.alternate_name
        self.model_generator.generate_single_model(
            model_name, model_shape=self.operation_model.output_shape
        )
        return model_name

    @property
    def response_attr(self) -> str | None:
        """
        Deduce the name of the attribute in the boto3 response that we want to
        return from the method.  This is either some variation of the name of
        the model itself, or whatever the botocraft config for the method
        specifies.

        Returns:
            The name of the attribute in the boto3 response that we want to
            return from the method.

        """
        response_attr: str | None = None
        if self.output_shape is None:
            return None
        if not hasattr(self.output_shape, "members"):
            return None
        if self.method_def.response_attr:
            if self.method_def.response_attr == "None":
                return None
            return self.method_def.response_attr
        potential_names = [
            self.model_name.lower(),
            self.model_name_plural.lower(),
        ]
        # We need to take into account when we rename an attribute on a response
        # shape, e.g. ``CreateLaunchTemplateResult`` has an attribute
        # ``LaunchTemplate`` that we rename to ``LaunchTemplateInstance``.
        # Searching through the shape.members dictionary doesn't work in this
        # case, because the attribute name is different
        response_attrs = {attr.lower(): attr for attr in self.output_shape.members}
        response_model_def = self.model_generator.get_model_def(self.output_shape.name)
        for field_data in response_model_def.fields.values():
            if field_data.rename:
                response_attrs[field_data.rename.lower()] = field_data.rename
        for attr, value in response_attrs.items():
            if attr in potential_names:
                response_attr = value
                break
        attrs = ", ".join([f'"{attr}"' for attr in self.output_shape.members])
        if response_attr:
            # What we're doing here is trying to figure out if the response_attr is
            # references a specific key or index in a list or dictionary.  If it does,
            # strip that off so we can see if the attribute is actually a field on the
            # response class.
            test_attr = re.split(r"\[|\.", response_attr)[0]
            if test_attr not in self.output_shape.members:
                msg = (
                    f"{self.model_name}Manager.{self.method_name}: {test_attr} "
                    f"is not a field on the response class {self.output_shape.name}: "
                    f"{attrs}"
                )
                raise KeyError(msg)
            return response_attr
        msg = (
            f"Can't deduce response attribute for response class "
            f"{self.output_shape.name}: {attrs}"
        )
        raise ValueError(msg)

    @property
    def response_attr_multiplicity(self) -> Literal["one", "many"]:
        """
        Determine if the response attribute is a list or not.

        Returns:
            ``'one'`` if the response attribute is not a list, ``'many'`` if it
            is a list.

        """
        if self.output_shape is None:
            return "one"
        if not hasattr(self.output_shape, "members"):
            return "one"
        if self.response_attr is None:
            return "one"
        # What we're doing here is trying to figure out if the response_attr is
        # references a specific key or index in a list or dictionary.  If it does,
        # strip that off so we can see if the attribute is actually a field on the
        # response class.
        test_attr = re.split(r"\[|\.", self.response_attr)[0]
        try:
            shape = self.output_shape.members[test_attr]
        except KeyError:
            response_model_def = self.model_generator.get_model_def(
                self.output_shape.name
            )
            for field, field_data in response_model_def.fields.items():
                if field_data.rename == self.response_attr:
                    shape = self.output_shape.members[field]
                    break
            else:
                raise
            # Maybe we're dealing with a renamed field
        if hasattr(shape, "type") and shape.type == "list":
            return "many"
        if hasattr(shape, "type_name") and shape.type_name == "list":
            return "many"
        return "one"

    @property
    def return_type(self) -> str:
        """
        Set the type hint for the return type of the method.  This is either the
        response class, or whatever class the botocraft config for the method
        specifies.

        Returns:
            The type hint for the return type of the method.

        """
        if self.method_def.return_type:
            return self.method_def.return_type
        # If our output shape has no members, then we return None
        output_shape = cast("botocore.model.StructureShape", self.output_shape)
        if not hasattr(output_shape, "members") or (
            hasattr(output_shape, "members") and not output_shape.members
        ):
            return "None"
        if self.response_attr is None:
            if self.output_shape is None:
                return "None"
            return self.shape_converter.convert(self.output_shape, quote=True)
        if self.output_shape is not None:
            try:
                response_attr_shape = self.output_shape.members[
                    cast("str", self.response_attr)
                ]
            except KeyError:
                response_model_def = self.model_generator.get_model_def(
                    self.output_shape.name
                )
                for field, field_data in response_model_def.fields.items():
                    if field_data.rename == self.response_attr:
                        response_attr_shape = self.output_shape.members[
                            cast("str", field)
                        ]
                        break
                else:
                    raise
        return_type = self.shape_converter.convert(response_attr_shape, quote=True)
        # Deal with the primary model itself having an alternate name
        if self.model_def.alternate_name:
            if f'"{self.model_name}"' in return_type:
                return_type = return_type.replace(
                    f'"{self.model_name}"', f'"{self.model_def.alternate_name}"'
                )
        return return_type

    @property
    def decorators(self) -> str | None:
        """
        Return the decorators for the method, if any.

        Returns:
            Decorators for the method, or ``None`` if there are none.

        """
        if not self.method_def.decorators:
            return None
        code = ""
        for decorator in self.method_def.decorators:
            code += f"    @{decorator.name}\n"
            if decorator.import_path:
                self.imports.add(
                    f"from {decorator.import_path} import {decorator.name}"
                )
        return code + "\n"

    @property
    def signature(self) -> str:
        """
        Create the method signature for the method.

        Example:

            .. code-block:: python

                def create(self, name: str, *, description: str) -> Model:

        Returns:
            The method signature for the method.

        """
        args = self.args()
        kwargs = self.kwargs()
        args_str = ", ".join([f"{arg}: {arg_type}" for arg, arg_type in args.items()])
        kwargs_str = ", ".join(
            [f"{arg}: {arg_type}" for arg, arg_type in kwargs.items()]
        )
        signature = f"    def {self.method_name}(self, "
        if args_str:
            signature += args_str
        if args_str and kwargs_str:
            signature += ", "
        if kwargs_str:
            signature += f"*, {kwargs_str}"
        signature += f") -> {self.return_type}:"
        return signature

    def get_arg_docstring(self, arg: str) -> str | None:
        """
        Return the docstring for the given argument.

        Args:
            arg: the name of the argument

        Returns:
            The docstring for the argument.

        """
        _arg_name = arg
        arg_def = MethodArgumentDefinition()
        for arg_name, _arg_def in self.method_def.args.items():
            if arg_name == arg:
                arg_def = _arg_def
                break
            if _arg_def.rename == arg:
                arg_def = _arg_def
                _arg_name = arg_name
                break
        if arg_def.docstring:
            return arg_def.docstring
        if self.input_shape is not None:
            if _arg_name in self.input_shape.members:
                return self.input_shape.members[_arg_name].documentation
        return None

    @property
    def docstrings_def(self) -> MethodDocstringDefinition:
        """
        Build the docstring definition for the method.

        Returns:
            A :py:class:`MethodDocstringDefinition` instance.

        """
        docstrings: MethodDocstringDefinition = MethodDocstringDefinition()
        docstrings.method = (
            self.method_def.docstring
            if self.method_def.docstring
            else self.operation_model.documentation
        )
        for arg in self.args():
            docstrings.args[arg] = self.get_arg_docstring(arg)
        for arg in self.kwargs():
            docstrings.kwargs[arg] = self.get_arg_docstring(arg)
        return docstrings

    @property
    def docstring(self) -> str | None:
        """
        Return the docstring for the method.
        """
        docs = self.docstrings_def.render(self.docformatter)
        if not docs:
            return None
        return re.sub(r"\\{1}", "", docs, flags=re.MULTILINE)

    @property
    def body(self) -> str:
        """
        Return the body of the method.  It's implemented by the subclasses.

        Returns:
            The body of the method.

        """
        raise NotImplementedError

    @property
    def code(self) -> str:
        """
        Return the full code for the method.

        Returns:
            The full code for the method, ready to be inserted into the
            generated manager class.

        """
        code = ""
        if decorators := self.decorators:
            code += decorators
        code += self.signature
        if docstring := self.docstring:
            code += docstring
        code += self.body
        return code



from typing import TYPE_CHECKING

import botocore

if TYPE_CHECKING:
    from botocraft.sync.shapes import AbstractShapeConverter


class ModelHasNoMembersError(Exception):
    """Exception raised when a model has no members."""

    def __init__(self, model_name: str):
        super().__init__(f"Model '{model_name}' has no fields.")
        self.model_name = model_name
        self.message = f"Model '{model_name}' has no fields."


class NoPythonTypeError(Exception):
    """Exception raised when a model has no Python type."""

    def __init__(self, field_name: str, model_name: str):
        super().__init__(
            f"Field '{field_name}' on model '{model_name}' has no Python type."
        )
        self.model_name = model_name
        self.field_name = field_name
        self.message = (
            f"Field '{field_name}' on model '{model_name}' has no Python type."
        )


class WrongConverterError(Exception):
    """Exception raised when a shape converter is not the right one for a shape."""

    def __init__(
        self, shape: botocore.model.Shape, converter: "AbstractShapeConverter"
    ):
        self.shape_name = shape.name
        self.converter_name = converter.__class__.__name__
        super().__init__(
            f"Shape '{self.shape_name}' cannot be converted by {self.converter_name}."
        )
        self.message = (
            f"Shape '{self.shape_name}' cannot be converted by {self.converter_name}."
        )


class NoConverterError(Exception):
    """Exception raised when a shape converter is not found."""

    def __init__(self, shape: botocore.model.Shape):
        self.shape_name = shape.name
        self.type_name = shape.type_name
        super().__init__(
            f"No converter found for shape '{self.shape_name}' [{self.type_name}]."
        )
        self.message = (
            f"No converter found for shape '{self.shape_name}' [{self.type_name}]."
        )
